"""
Model Generator for OpenSees ~ collections
Collections are designated containers of objects of a particular type.
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import Any
from typing import TypeVar
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from . import common
from .ops import node
from .ops import element


nparr = npt.NDArray[np.float64]

TK = TypeVar('TK')
TV = TypeVar('TV')


@dataclass(repr=False)
class Collection(dict[TK, TV]):
    """
    Collection of objects.
    Attributes:
        parent (Any)
    """
    parent: Any = field(repr=False)

    def add(self, obj):
        """
        Add an object to the collection
        """
        if not hasattr(obj, 'uid'):
            raise KeyError('Object does not have a uid attribute')
        if obj.uid in self:
            raise KeyError('Object uid already exists')
        self[obj.uid] = obj

    def retrieve_by_attr(self, attr: Any, val: Any) -> Any:
        """
        Retrieve an object from the collection based on an attribute
        value
        """
        res = None
        for thing in self.values():
            if hasattr(thing, attr):
                other_val = getattr(thing, attr)
                if other_val == val:
                    res = thing
        return res

    def __srepr__(self):
        """
        Short version of repr
        """
        return f'[Collection of {len(self)} items]'

    def __repr__(self):
        res = ''
        res += 'Collection Object\n'
        res += f'ID: {id(self)}\n'
        res += f'Parent object: {self.parent}\n'
        res += f'Registry size: {len(self)}\n'
        return res


@dataclass(repr=False)
class CollectionActive(Collection[TK, TV]):
    """
    Collection with support for currently active objects
    """
    active: list[TK] = field(default_factory=list)

    def set_active(self, uids: list[TK]):
        """
        Sets the active objects.
        Args:
            uids (list[int]): uids of the objects to set as active
        """
        for uid in uids:
            assert uid in self
        self.active = uids

    def set_active_all(self):
        """
        Sets the active objects.
        Args:
            uids (list[int]): uids of the objects to set as active
        """
        self.active = []
        for key in self:
            self.active.append(key)


@dataclass(repr=False)
class NodeCollection(Collection[int, node.Node]):
    """
    Node collection.
    Attributes:
        parent (Any)
    """
    named_contents: dict[str, node.Node] = field(default_factory=dict)

    def search_xy(self, x_coord, y_coord):
        """
        Returns the node that occupies a given point if it exists
        """

        candidate_pt: nparr = np.array(
            [x_coord, y_coord, self.parent.elevation])
        for other_node in self.values():
            other_pt: nparr = np.array(other_node.coords)
            if np.linalg.norm(candidate_pt - other_pt) < common.EPSILON:
                return other_node
        # indent the following line once to the right, and you'll be
        # spending a couple of nights trying to figure out why there
        # are so many duplicate nodes
        return None


@dataclass(repr=False)
class CollectionWithConnectivity(Collection[TK, TV]):
    """
    Collection of elements for which it is important to consider their
    connectivity.
    Attributes:
        parent (Any)
    """
    named_contents: dict[
        str, element.ElasticBeamColumn] = field(
            default_factory=dict)

    def add(self, elm):
        """
        Adds an element to the collection.
        The method also checks to see if an object having the same
        connectivity exists in the collection, and raises an error if
        it does.
        """
        uids = [nd.uid for nd in elm.nodes]
        uids.sort()
        uids_tuple = (*uids,)
        if uids_tuple in elm.parent_component.element_connectivity():
            raise ValueError('This should never happen!')
        super().add(elm)



"""
Generates meshes for preconfigured sections
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import numpy as np
import numpy.typing as npt
from ..mesh import Mesh
from ..mesh import Vertex
from ..mesh import Edge
from ..mesh import orient_loops
from ..mesh import obtain_closed_loops
from ..mesh import sanity_checks
from ..mesh import define_halfedges
from ..import common

nparr = npt.NDArray[np.float64]


def generate(edges):
    """
    Generates halfedges from the given edges
    """
    halfedges = define_halfedges(edges)
    loops = obtain_closed_loops(halfedges)
    _, internal, trivial = orient_loops(loops)
    sanity_checks(internal, trivial)
    return Mesh(internal[0])


def define_edges(vertices):
    """
    Defines edges from an ordered list of vertices
    """
    n_v = len(vertices)
    edges = []
    for i in range(n_v - 1):
        v_i = vertices[i]
        v_j = vertices[i+1]
        edges.append(Edge(v_i, v_j))
    v_i = vertices[-1]
    v_j = vertices[0]
    edges.append(Edge(v_i, v_j))
    return edges


def w_mesh(sec_b, sec_h, sec_tw, sec_tf, target_area=None):
    """
    Defines a loop of counterclockwise halfedges
    that form the shape of the W section with
    the specified parameters.
    The origin coincides with the centroid.
    Input:
        b: total width
        h: total height
        tw: web thickness
        tf: flange thickness
        target_area: AISC database area
            to determine fillets,
            because trying to do that using
            `T` doesn't work.
    """
    area_diff = target_area - (sec_b*sec_tf*2.+(sec_h-2*sec_tf)*sec_tw)
    if area_diff < 0:
        # This happens for W14X426
        area_diff = 1e-4

    dist = np.sqrt(area_diff/(2.**2-np.pi)) * 0.9565
    # note: 0.9565 is a correction factor to account for
    # the fact that we approximate the arcs with
    # four line segments, thus putting more material in there
    k = (sec_b - 2. * dist - sec_tw) / 2.
    vertices = [
        Vertex((sec_b/2., sec_h/2.)),
        Vertex((-sec_b/2., sec_h/2.)),
        Vertex((-sec_b/2., sec_h/2.-sec_tf)),
        Vertex((-sec_b/2.+k, sec_h/2.-sec_tf)),
        Vertex((-sec_b/2.+k+dist * np.cos(3.*np.pi/8.),
                sec_h/2.-sec_tf-dist + dist*np.sin(3.*np.pi/8.))),
        Vertex((-sec_b/2.+k+dist*np.cos(1.*np.pi/4.),
                sec_h/2.-sec_tf-dist+dist*np.sin(1.*np.pi/4.))),
        Vertex((-sec_b/2.+k+dist*np.cos(1.*np.pi/8.),
                sec_h/2.-sec_tf-dist+dist*np.sin(1.*np.pi/8.))),

        Vertex((-sec_b/2.+k+dist, sec_h/2.-sec_tf-dist)),
        Vertex((-sec_b/2.+k+dist, -sec_h/2.+sec_tf+dist)),
        Vertex((-sec_b/2.+k+dist*np.cos(1.*np.pi/8.),
                -sec_h/2.+sec_tf+dist-dist*np.sin(1.*np.pi/8.))),
        Vertex((-sec_b/2.+k+dist*np.cos(1.*np.pi/4.),
                -sec_h/2.+sec_tf+dist-dist*np.sin(1.*np.pi/4.))),
        Vertex((-sec_b/2.+k+dist*np.cos(3.*np.pi/8.),
                -sec_h/2.+sec_tf+dist-dist*np.sin(3.*np.pi/8.))),

        Vertex((-sec_b/2.+k, -sec_h/2.+sec_tf)),
        Vertex((-sec_b/2., -(sec_h/2.-sec_tf))),
        Vertex((-sec_b/2., -sec_h/2.)),
        Vertex((sec_b/2., -sec_h/2.)),
        Vertex((sec_b/2., -(sec_h/2-sec_tf))),
        Vertex((+sec_b/2.-k, -sec_h/2.+sec_tf)),
        Vertex((+sec_b/2.-k-dist*np.cos(3.*np.pi/8.),
                -sec_h/2.+sec_tf+dist-dist*np.sin(3.*np.pi/8.))),
        Vertex((+sec_b/2.-k-dist*np.cos(1.*np.pi/4.),
                -sec_h/2.+sec_tf+dist-dist*np.sin(1.*np.pi/4.))),
        Vertex((+sec_b/2.-k-dist*np.cos(1.*np.pi/8.),
                -sec_h/2.+sec_tf+dist-dist*np.sin(1.*np.pi/8.))),

        Vertex((+sec_b/2.-k-dist, -sec_h/2.+sec_tf+dist)),
        Vertex((+sec_b/2.-k-dist, +sec_h/2.-sec_tf-dist)),
        Vertex((+sec_b/2.-k-dist*np.cos(1.*np.pi/8.),
                +sec_h/2.-sec_tf-dist+dist*np.sin(1.*np.pi/8.))),
        Vertex((+sec_b/2.-k-dist*np.cos(1.*np.pi/4.),
                +sec_h/2.-sec_tf-dist+dist*np.sin(1.*np.pi/4.))),
        Vertex((+sec_b/2.-k-dist*np.cos(3.*np.pi/8.),
                +sec_h/2.-sec_tf-dist+dist*np.sin(3.*np.pi/8.))),

        Vertex((+sec_b/2.-k, sec_h/2.-sec_tf)),
        Vertex((sec_b/2., sec_h/2.-sec_tf))
    ]
    edges = define_edges(vertices)
    return generate(edges)


def hss_rect_mesh(sec_ht: float, sec_b: float, sec_t: float):
    """
    Defines a loop of counterclockwise halfedges
    that form the shape of the rectangular HSS
    with the specified parameters.
    The origin coincides with the centroid.
    Input:
        ht (float): Overall depth
        b (float): Overall width
        t (float): Wall thickness
    """
    dim_a = sec_b / 2.
    dim_c = sec_ht / 2.
    dim_u = dim_a - sec_t
    dim_v = dim_c - sec_t
    dim_e = common.EPSILON
    vertices = [
        Vertex((+dim_e, -dim_c)),
        Vertex((+dim_a, -dim_c)),
        Vertex((+dim_a, +dim_c)),
        Vertex((-dim_a, +dim_c)),
        Vertex((-dim_a, -dim_c)),
        Vertex((-dim_e, -dim_c)),
        Vertex((-dim_e, -dim_v)),
        Vertex((-dim_u, -dim_v)),
        Vertex((-dim_u, +dim_v)),
        Vertex((+dim_u, +dim_v)),
        Vertex((+dim_u, -dim_v)),
        Vertex((+dim_e, -dim_v)),
    ]
    edges = define_edges(vertices)
    return generate(edges)


def hss_circ_mesh(o_d: float, t_des: float, n_pts: int):
    """
    Defines a loop of counterclockwise halfedges
    that form the shape of the circular
    HSS with the specified parameters.
    The origin coincides with the centroid.
    Input:
        o_d (float): Outside diameter
        t_des (float): Design wall thickness
        n_pts (int) Number of points to approximate
              a circle.
    """
    dim_e = common.EPSILON
    t_param = np.linspace(0., 2.*np.pi, n_pts)
    pts_normalized: nparr = np.column_stack(
        (np.sin(t_param), -np.cos(t_param)))
    pts_outer = pts_normalized * o_d
    pts_outer[0, 0] += dim_e
    pts_outer[-1, 0] -= dim_e
    pts_inner: nparr = np.flip(pts_normalized * (o_d - t_des), axis=0)
    pts_inner[0, 0] -= dim_e
    pts_inner[-1, 0] += dim_e
    pts_all: nparr = np.concatenate((pts_outer, pts_inner))
    vertices = []
    for point in pts_all:
        vertices.append(Vertex((point[0], point[1])))
    edges = define_edges(vertices)
    return generate(edges)


def rect_mesh(dim_b, dim_h):
    """
    Defines a loop of counterclockwise halfedges
    that form the shape of the rectangular section with
    the specified parameters.
    The origin coincides with the centroid.
    Input:
        b: total width
        h: total height
    """
    vertices = [
        Vertex((dim_b/2., dim_h/2.)),
        Vertex((-dim_b/2., dim_h/2.)),
        Vertex((-dim_b/2., -dim_h/2.)),
        Vertex((dim_b/2., -dim_h/2.))
    ]
    edges = define_edges(vertices)
    return generate(edges)


def generic_snap_points(mesh: Mesh) -> dict[str, nparr]:
    """
    Generates generic snap poitns
    for a section object.
    """
    bbox = mesh.bounding_box()
    z_min, y_min, z_max, y_max = bbox.flatten()
    snap_points: dict[str, nparr] = {}
    snap_points['centroid'] = - np.array([0., 0.])
    snap_points['top_center'] = - np.array([0., y_max])
    snap_points['top_left'] = - np.array([z_min, y_max])
    snap_points['top_right'] = - np.array([z_max, y_max])
    snap_points['center_left'] = - np.array([z_min, 0.])
    snap_points['center_right'] = - np.array([z_max, 0.])
    snap_points['bottom_center'] = - np.array([0., y_min])
    snap_points['bottom_left'] = - np.array([z_min, y_min])
    snap_points['bottom_right'] = - np.array([z_max, y_min])
    return snap_points



"""
Model Generator for OpenSees ~ split component
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import numpy as np
from ..line import Line
from ..ops.element import ElasticBeamColumn
from ..ops.element import DispBeamColumn
from ..ops.element import GeomTransf
from ..ops.node import Node
from .. import common
from ..ops.element import Lobatto


def split_component(component, point):
    """
    Splits a beam-functioning component assembly to
    accomodate for a node required to connect
    another component assembly
    """
    elms = []
    elms.extend(component.elastic_beamcolumn_elements.values())
    elms.extend(component.disp_beamcolumn_elements.values())
    distances = np.zeros(len(elms))
    for i, elm in enumerate(elms):
        p_i = (np.array(elm.nodes[0].coords)
               + elm.geomtransf.offset_i)
        p_j = (np.array(elm.nodes[1].coords)
               + elm.geomtransf.offset_j)
        line = Line('', p_i, p_j)
        dist = line.point_distance(point)
        distances[i] = dist
    np.nan_to_num(distances, copy=False, nan=np.inf)
    i_min = np.argmin(distances)
    closest_elm = elms[i_min]
    p_i = (np.array(closest_elm.nodes[0].coords)
           + closest_elm.geomtransf.offset_i)
    p_j = (np.array(closest_elm.nodes[1].coords)
           + closest_elm.geomtransf.offset_j)
    line = Line('', p_i, p_j)
    split_point = line.project(point)
    assert split_point is not None  # check if it exists

    # first check if a node already exists there
    inodes = component.internal_nodes.values()
    for inode in inodes:
        if np.linalg.norm(
                np.array(inode.coords) - split_point) < common.EPSILON:
            avail_node = inode
            offset = point - np.array(avail_node.coords)
            return avail_node, offset

    # otherwise:

    # remove existing line element
    node_i = closest_elm.nodes[0]
    node_j = closest_elm.nodes[1]
    prev_section = closest_elm.section
    prev_gtransf = closest_elm.geomtransf
    if isinstance(closest_elm, ElasticBeamColumn):
        component.elastic_beamcolumn_elements.pop(closest_elm.uid)
    elif isinstance(closest_elm, DispBeamColumn):
        component.disp_beamcolumn_elements.pop(closest_elm.uid)
    else:
        raise ValueError('Unsupported element type')

    # add split node
    middle_node = Node(
        component.parent_collection.parent
        .parent_model.uid_generator.new('node'),
        list(split_point))
    component.internal_nodes.add(middle_node)
    # add two new line elements
    # part i
    o_i = prev_gtransf.offset_i
    o_j = np.zeros(3)
    n_i = node_i
    n_j = middle_node
    transf_i = GeomTransf(
        prev_gtransf.transf_type,
        component.parent_collection.parent.parent_model
        .uid_generator.new('transformation'),
        o_i,
        o_j,
        prev_gtransf.x_axis,
        prev_gtransf.y_axis,
        prev_gtransf.z_axis
    )
    if isinstance(closest_elm, ElasticBeamColumn):
        elm_i = ElasticBeamColumn(
            component,
            component.parent_collection.parent.parent_model
            .uid_generator.new('element'),
            [n_i, n_j],
            prev_section,
            transf_i
        )
        component.elastic_beamcolumn_elements.add(elm_i)
    elif isinstance(closest_elm, DispBeamColumn):
        assert isinstance(closest_elm.integration, Lobatto)
        beam_integration = Lobatto(
            uid=component.parent_collection.parent.parent_model
            .uid_generator.new('beam integration'),
            parent_section=prev_section,
            n_p=closest_elm.integration.n_p
        )
        elm_i = DispBeamColumn(  # type: ignore
            component,
            component.parent_collection.parent.parent_model
            .uid_generator.new('element'),
            [n_i, n_j],
            prev_section,
            transf_i,
            beam_integration
        )
        component.disp_beamcolumn_elements.add(elm_i)
    # part j
    o_i = np.zeros(3)
    o_j = prev_gtransf.offset_j
    n_i = middle_node
    n_j = node_j
    transf_j = GeomTransf(
        prev_gtransf.transf_type,
        component.parent_collection.parent.parent_model
        .uid_generator.new('transformation'),
        o_i,
        o_j,
        prev_gtransf.x_axis,
        prev_gtransf.y_axis,
        prev_gtransf.z_axis
    )
    if isinstance(closest_elm, ElasticBeamColumn):
        elm_j = ElasticBeamColumn(
            component,
            component.parent_collection.parent.parent_model
            .uid_generator.new('element'),
            [n_i, n_j],
            prev_section,
            transf_j
        )
        component.elastic_beamcolumn_elements.add(elm_j)
    elif isinstance(closest_elm, DispBeamColumn):
        assert isinstance(closest_elm.integration, Lobatto)
        beam_integration = Lobatto(
            uid=component.parent_collection.parent.parent_model.
            uid_generator.new('beam integration'),
            parent_section=prev_section,
            n_p=closest_elm.integration.n_p
        )
        elm_j = DispBeamColumn(  # type: ignore
            component,
            component.parent_collection.parent.parent_model
            .uid_generator.new('element'),
            [n_i, n_j],
            prev_section,
            transf_j,
            beam_integration
        )
        component.disp_beamcolumn_elements.add(elm_j)

    # calculate offset and return
    offset = point - np.array(middle_node.coords)
    return middle_node, offset



"""
Model Generator for OpenSees ~ level
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass, field
from .collections import Collection
from .collections import NodeCollection
if TYPE_CHECKING:
    from .model import Model
    from .component_assembly import ComponentAssembly


@dataclass
class Level:
    """
    Level Object
    Levels are part of a model and they contain primary nodes and
      component assemblies.
    Attributes:
        parent_model (Model)
        uid (int)
        elevation (float)
        nodes (NodeCollection)
        components (Collection)
    """
    parent_model: Model = field(repr=False)
    uid: int
    elevation: float
    nodes: NodeCollection = field(init=False, repr=False)
    components: Collection[
        int, ComponentAssembly] = field(init=False, repr=False)

    def __post_init__(self):
        self.nodes = NodeCollection(self)
        self.components = Collection(self)

    def __repr__(self):
        res = ''
        res += 'Level object\n'
        res += f'parent_model: {self.parent_model.name}\n'
        res += f'uid: {self.uid}\n'
        res += f'elevation: {self.elevation}\n'
        res += 'Nodes: \n'
        res += self.nodes.__srepr__() + '\n'
        res += 'Components: \n'
        res += self.components.__srepr__() + '\n'
        return res



"""
Model Generator for OpenSees ~ physical material
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from dataclasses import dataclass


@dataclass
class PhysicalMaterial:
    """
    Physical material.
    We use this for self-weight, plotting enhancements etc.
    """
    uid: int
    name: str
    variety: str
    density: float
    e_mod: float
    g_mod: float
    f_y: float

    def __srepr__(self):
        """
        Short version of repr
        """
        return f'Physical material: {self.name}, {self.uid}'

    def __repr__(self):
        res = ''
        res += 'Physical material object\n'
        res += f'Name: {self.name}\n'
        res += f'uid: {self.uid}\n'
        res += f'variety: {self.variety}\n'
        res += f'density: {self.density}\n'
        res += f'E: {self.e_mod}\n'
        res += f'G: {self.g_mod}\n'
        return res



"""
Common utility functions used for 3D data visualization
https://plotly.com/python/reference/
"""
#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import plotly.graph_objects as go  # type: ignore
# pio.renderers.default = 'browser'

# # used for production plots
# node_marker = {
#     'fixed': ("square", 8),
#     'pinned': ("circle-open", 2),
#     'free': ("circle", 3),
#     'parent': ("circle-open", 15),
#     'internal': ("x", 2),
#     '111011': ("circle-open", 3),
# }

# used for development plots
node_marker = {
    'fixed': ("square", 10),
    'free': ("circle", 5),
    'parent': ("circle-open", 20),
    'internal': ("x", 3),
    'release': ('circle-open', 15)
}


def global_layout(camera=None):
    """
    Some general definitions needed often
    """
    # view_type = "orthographic"
    view_type = "perspective"
    if not camera:
        camera = dict(
                       up=dict(x=0, y=0, z=1),
                       center=dict(x=0, y=0, z=0),
                       eye=dict(x=1.25, y=1.25, z=1.25),
                       projection={
                           "type": view_type
                       }
                   )
    return go.Layout(
        scene=dict(aspectmode='data',
                   xaxis_visible=False,
                   yaxis_visible=False,
                   zaxis_visible=False,
                   bgcolor='black',
                   camera=camera
                   )
    )


if __name__ == "__main__":
    pass



"""
Library for consistent all-purpose 2D plots
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

#############
# Resources #
#############
#
# https://coolors.co/  <- amazing color palette generator
#

import sys
import numpy as np
import numpy.typing as npt
import plotly.graph_objects as go

nparr = npt.NDArray[np.float64]


C_LINE = "#6F1D1B"
C_AXIS_LINE = "#432818"
C_TITLE_COLOR = "#432818"
C_LABEL_COLOR = "#432818"
C_GRID = "#BB9457"
C_BACKGROUND = "#FFE6A7"
C_HOVERLABEL_BG = "#BB9457"


def line_plot_interactive(title_text, x_vals, y_vals, mode,
                          xlab="x", xunit=None, xhoverformat='.0f',
                          ylab="y", yunit=None, yhoverformat='.0f'):
    """
    Gneric polty line plot
    """
    assert len(x_vals) == len(y_vals), "Dimensions don't match"
    assert mode in ['spline+markers', 'line'], \
        "mode can either be `spline+markers` or `line`"

    if mode == 'line':
        lshape = 'linear'
        lmode = 'lines'
        lwidth = 2
    elif mode == 'spline+markers':
        lshape = 'spline'
        lmode = 'lines+markers'
        lwidth = 4
    else:
        raise ValueError("oops! This should never run. Strange.")

    if xunit:
        xtitle = xlab + ' (' + xunit + ')'
    else:
        xtitle = xlab
    if yunit:
        ytitle = ylab + ' (' + yunit + ')'
    else:
        ytitle = ylab

    num_points = len(x_vals)
    indices: nparr = np.array([range(num_points)])
    my_hovertemplate = \
        'XY value pair: %{customdata:d}<br>' + \
        xlab + ' = %{x: ' + xhoverformat + '} '
    if xunit:
        my_hovertemplate += xunit + '<br>'
    else:
        my_hovertemplate += '<br>'
    my_hovertemplate += ylab+' = %{y:' + yhoverformat + '} '
    if yunit:
        my_hovertemplate += yunit + '<extra></extra>'
    else:
        my_hovertemplate += '<extra></extra>'
    fig = go.Figure(
        data=go.Scatter(
            x=x_vals,
            y=y_vals,
            mode=lmode,
            line_shape=lshape,
            line=dict(color=C_LINE, width=lwidth),
            marker=dict(color=C_BACKGROUND,
                        size=10,
                        line=dict(
                            color=C_LINE,
                            width=4)),
            customdata=indices,
            hovertemplate=my_hovertemplate
        )
    )
    fig.update_layout(
        xaxis=dict(
            showline=True,
            showgrid=True,
            gridcolor=C_GRID,
            showticklabels=True,
            linecolor=C_BACKGROUND,
            zerolinecolor=C_AXIS_LINE,
            linewidth=0,
            ticks='outside',
            tickcolor=C_GRID,
            tickfont=dict(
                family='Cambria',
                size=14,
                color=C_AXIS_LINE,
            ),
        ),
        yaxis=dict(
            showline=True,
            showgrid=True,
            gridcolor=C_GRID,
            showticklabels=True,
            linecolor=C_BACKGROUND,
            zerolinecolor=C_AXIS_LINE,
            linewidth=0,
            ticks='outside',
            tickcolor=C_GRID,
            tickfont=dict(
                family='Cambria',
                size=14,
                color=C_AXIS_LINE,
            ),
        ),
        showlegend=False,
        plot_bgcolor=C_BACKGROUND
    )
    annotations = [
        dict(
            xref='paper',
            yref='paper',
            x=0.0,
            y=1.00,
            xanchor='left', yanchor='bottom',
            text=title_text,
            font=dict(family='Cambria',
                      size=28,
                      color=C_TITLE_COLOR),
            showarrow=False
        )
    ]
    fig.update_layout(annotations=annotations)
    fig.update_layout(
        xaxis_title=xtitle,
        yaxis_title=ytitle,
        font=dict(
            family='Cambria',
            size=18,
            color=C_LABEL_COLOR
        ),
        hoverlabel=dict(
            bgcolor=C_HOVERLABEL_BG,
            font_size=15,
            font_family='Cambria'
        )
    )
    if "pytest" not in sys.modules:
        fig.show()


if __name__ == '__main__':
    pass


# # (Test)
# x = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
# y = [0, 10, 20, 30, 35, 38, 40, 41, 42, 43]

# line_plot_interactive('Pushover\nDirection: Y',
#                       'Displacement', 'Base Shear',
#                       'in', 'lb', x, y)



"""
Model Generator for OpenSees ~ self weight, self mass
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import numpy as np
from .. import common


def self_weight(mdl, lcase, factor=1.00):
    """
    Assigns the structure's self weight to its members
    """
    for elm in mdl.list_of_beamcolumn_elements():

        # if mdl.settings.imperial_units:
        #     g_const = common.G_CONST_IMPERIAL
        # else:
        #     g_const = common.G_CONST_SI

        weight_per_length = elm.section.weight_per_length()
        # apply weight as UDL
        if elm.visibility.skip_opensees_definition:
            # in that case apply its weight to the connecting nodes
            elm_len = elm.clear_length()
            elm_w = weight_per_length * elm_len * factor
            lcase.node_loads[
                elm.nodes[0].uid].add(
                    [0.00, 0.00, -elm_w/2.00, 0.00, 0.00, 0.00])
            lcase.node_loads[
                elm.nodes[1].uid].add(
                    [0.00, 0.00, -elm_w/2.00, 0.00, 0.00, 0.00])
        else:
            lcase.line_element_udl[
                elm.uid].add_glob(np.array(
                    [0., 0., -weight_per_length*factor]))


def self_mass(mdl, lcase):
    """
    Assigns the structure's self mass to its members
    """
    for elm in mdl.list_of_beamcolumn_elements():

        if mdl.settings.imperial_units:
            g_const = common.G_CONST_IMPERIAL
        else:
            g_const = common.G_CONST_SI

        weight_per_length = elm.section.weight_per_length()
        mass_per_length = weight_per_length / g_const
        # apply lumped mass at the connecting nodes
        half_mass = (mass_per_length *
                     elm.clear_length() / 2.00)
        lcase.node_mass[
            elm.nodes[0].uid].add([half_mass]*3+[0.00]*3)
        lcase.node_mass[
            elm.nodes[1].uid].add([half_mass]*3+[0.00]*3)



"""
Model Generator for OpenSees ~ design library
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
import pandas as pd
from ..solver import Analysis
from .basic_forces import basic_forces
from ..model import Model
from ..solver import ModalResponseSpectrumAnalysis

nparr = npt.NDArray[np.float64]


CaseName = str
ComponentName = str


@dataclass(repr=False)
class LoadCombination:
    """
    Load combinations
    stuff in the first list are added.
    stuff inside the sub-list are enveloped
    """
    mdl: Model
    combo: dict[ComponentName,
                list[tuple[float, Analysis, CaseName]]] = \
        field(default_factory=dict)

    def envelope_basic_forces(self, elm, num_points):
        """
        Calculates the envelope of the basic forces for the
        given load combination
        """
        df_min = pd.DataFrame(
            np.full((num_points, 6), np.inf),
            columns=['nx', 'qy', 'qz', 'tx', 'mz', 'my'])
        df_max = pd.DataFrame(
            np.full((num_points, 6), -np.inf),
            columns=['nx', 'qy', 'qz', 'tx', 'mz', 'my'])
        for component_to_envelope in self.combo.values():
            df_tot = pd.DataFrame(
                np.full((num_points, 6), 0.00),
                columns=['nx', 'qy', 'qz', 'tx', 'mz', 'my'])
            for component_to_add in component_to_envelope:
                factor, anl, case_name_str = component_to_add
                res = basic_forces(
                    anl, case_name_str, 0, elm, num_points)
                assert isinstance(res, pd.DataFrame)
                dframe = res * factor
                df_tot += dframe
            df_min[df_min > df_tot] = df_tot
            df_max[df_max < df_tot] = df_tot

        return df_min, df_max

    def envelope_node_displacement(self, node):
        """
        Calculates the enveloped node displacement
        """
        disp_min: nparr = np.full(6, np.inf)
        disp_max: nparr = np.full(6, -np.inf)
        for component_to_envelope in self.combo.values():
            disp_tot = np.full(6, 0.00)
            for component_to_add in component_to_envelope:
                factor, anl, case_name_str = component_to_add
                if isinstance(anl, ModalResponseSpectrumAnalysis):
                    disp: nparr = np.array(
                        anl.combined_node_disp(node.uid) * factor)
                else:
                    disp = np.array(anl.results[case_name_str]
                                    .node_displacements[node.uid][0])
                disp_tot += disp
            disp_min[disp_min > disp_tot] = disp[disp_min > disp_tot]
            disp_max[disp_max < disp_tot] = disp[disp_max < disp_tot]

        return disp_min, disp_max

    def envelope_node_displacement_diff(self, node_i, node_j):
        """
        Calculates the enveloped displacement difference between
        two nodes
        """
        disp_min = np.full(6, np.inf)
        disp_max = np.full(6, -np.inf)
        for component_to_envelope in self.combo.values():
            disp_tot = np.full(6, 0.00)
            for component_to_add in component_to_envelope:
                factor, anl, case_name_str = component_to_add
                if isinstance(anl, ModalResponseSpectrumAnalysis):
                    disp: nparr = np.array(anl.combined_node_disp_diff(
                        node_i.uid, node_j.uid) * factor)
                else:
                    disp_i: nparr = np.array(
                        anl.results[case_name_str]
                        .node_displacements[node_i.uid][0])
                    disp_j: nparr = np.array(
                        anl.results[case_name_str]
                        .node_displacements[node_j.uid][0])
                    disp = disp_i - disp_j
                disp_tot += disp
            disp_min[disp_min > disp_tot] = disp[disp_min > disp_tot]
            disp_max[disp_max < disp_tot] = disp[disp_max < disp_tot]

        return disp_tot, disp_tot



"""
Model Generator for OpenSees ~ plain beamcolumn element generator
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator


# pylint: disable=dangerous-default-value

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Type
from typing import Union
from typing import Callable
from dataclasses import dataclass, field
import numpy as np
import numpy.typing as npt
from ..ops.node import Node
from ..component_assembly import ComponentAssembly
from .querry import ElmQuerry
from .node_gen import NodeGenerator
from ..ops.element import ElasticBeamColumn
from ..ops.element import DispBeamColumn
from ..ops.element import ZeroLength
from ..ops.element import GeomTransf
from ..ops.element import Lobatto
from ..ops.section import ElasticSection
from ..ops.section import FiberSection
from ..preprocessing.split_component import split_component
from ..transformations import local_axes_from_points_and_angle
from ..transformations import transformation_matrix
from ..defaults import load_util_rigid_elastic
from ..gen.zerolength_gen import steel_w_col_pz


if TYPE_CHECKING:
    from ..model import Model
    from ..level import Level


nparr = npt.NDArray[np.float64]


def retrieve_snap_pt_global_offset(placement, section, p_i, p_j, angle):
    """
    Returns the necessary offset to connect an element at a specified
    snap point of the section
    """
    if section.snap_points and (placement != 'centroid'):
        # obtain offset from section (local system)
        d_z, d_y = section.snap_points[placement]
        sec_offset_local: nparr = np.array([0.00, d_y, d_z])
        # retrieve local coordinate system
        x_axis, y_axis, z_axis = \
            local_axes_from_points_and_angle(
                p_i, p_j, angle)
        t_glob_to_loc = transformation_matrix(
            x_axis, y_axis, z_axis)
        t_loc_to_glob = t_glob_to_loc.T
        sec_offset_global = t_loc_to_glob @ sec_offset_local
    else:
        sec_offset_global = np.zeros(3)
    return sec_offset_global


def beam_placement_lookup(
        x_coord, y_coord, querry, ndg, lvls, key,
        user_offset, section_offset, split_existing, snap):
    """
    Performs lookup operations before placing a beam-functioning
    component assembly to determine how to connect it with
    respect to the other existing objects in the model.
    """
    lvl = lvls[key]
    node = querry.search_node_lvl(x_coord, y_coord, lvl.uid)
    pinit = np.array((x_coord, y_coord, lvl.elevation)) + user_offset
    e_o = user_offset.copy() + section_offset
    if not node:
        if split_existing:
            node, offset = split_component(split_existing, pinit)
            e_o += offset
        else:
            node = ndg.add_node_lvl(x_coord, y_coord, key)
    else:
        # first check if a panel zone or other type of joint-like
        # component assembly exists at that node
        result_node = None
        components = querry.retrieve_components_from_nodes([node], lvl.uid)
        for component in components.values():
            if component.component_purpose == 'steel_W_panel_zone':
                if snap in ['middle_front', 'middle_back']:
                    result_node = component.external_nodes.named_contents[snap]
                    e_o += np.array(
                        (0.00, 0.00, node.coords[2] - result_node.coords[2]))
                    node = result_node
                    return node, e_o
                if snap in [
                        'centroid', 'top_center', 'top_left',
                        'top_right', 'center_left', 'center_right',
                        'bottom_center', 'bottom_left', 'bottom_right']:
                    elm = component.elastic_beamcolumn_elements.named_contents[
                        'elm_interior']
                    d_z, d_y = elm.section.snap_points[snap]
                    sec_offset_local: nparr = -np.array([0.00, d_y, d_z])
                    # retrieve local coordinate system
                    x_axis = elm.geomtransf.x_axis
                    y_axis = elm.geomtransf.y_axis
                    z_axis = elm.geomtransf.z_axis
                    t_glob_to_loc = transformation_matrix(
                        x_axis, y_axis, z_axis)
                    t_loc_to_glob = t_glob_to_loc.T
                    sec_offset_global = (
                        t_loc_to_glob @ sec_offset_local)
                    result_node = node
                    e_o += sec_offset_global
                    return node, e_o
                raise ValueError(f'Unsupported snap keyword: {snap}')

        # else check if a column-like component assembly exists
        if key-1 in lvls:
            node_below = querry.search_node_lvl(
                x_coord, y_coord, key-1)
            if node_below:
                column = querry.search_connectivity(
                    [node, node_below])
                if column:
                    elms = []
                    for dctkey in column.element_connectivity().keys():
                        if node.uid in dctkey:
                            elms.append(
                                column
                                .element_connectivity()[dctkey])
                    assert elms, 'There should be an element here.'
                    assert len(elms) == 1, \
                        'There should only be one element here.'
                    elm = elms[0]
                    # obtain offset from section (local system)
                    if elm.section.snap_points:
                        d_z, d_y = elm.section.snap_points[snap]
                        sec_offset_local = -np.array([0.00, d_y, d_z])
                        # retrieve local coordinate system
                        x_axis = elm.geomtransf.x_axis
                        y_axis = elm.geomtransf.y_axis
                        z_axis = elm.geomtransf.z_axis
                        t_glob_to_loc = transformation_matrix(
                            x_axis, y_axis, z_axis)
                        t_loc_to_glob = t_glob_to_loc.T
                        sec_offset_global = (
                            t_loc_to_glob @ sec_offset_local)
                        e_o += sec_offset_global
        else:
            raise ValueError(
                'Error: existing node without any elements to connect to.')
    return node, e_o


def look_for_panel_zone(
        node: Node,
        lvl: Level,
        querry: ElmQuerry
) -> Node:
    """
    Determines if a panel zone joint component assembly is present
    at the specified node.
    """
    components = querry.retrieve_components_from_nodes([node], lvl.uid)
    result_node = node
    for component in components.values():
        if component.component_purpose == 'steel_W_panel_zone':
            result_node = (component.external_nodes
                           .named_contents['bottom_node'])
            break
    return result_node


@dataclass(repr=False)
class BeamColumnGenerator:
    """
    This object introduces beamcolumn elements to a model.
    """
    model: Model = field(repr=False)

    def define_beamcolumn(
            self,
            assembly: ComponentAssembly,
            node_i: Node,
            node_j: Node,
            offset_i: nparr,
            offset_j: nparr,
            transf_type: str,
            section: ElasticSection | FiberSection,
            element_type: Type[Union[ElasticBeamColumn, DispBeamColumn]],
            angle=0.00) -> ElasticBeamColumn | DispBeamColumn:
        """
        Adds a beamcolumn element to the model, connecting the specified nodes
        """

        p_i = np.array(node_i.coords) + offset_i
        p_j = np.array(node_j.coords) + offset_j
        axes = local_axes_from_points_and_angle(
            p_i, p_j, angle)  # type: ignore
        if element_type.__name__ == 'ElasticBeamColumn':
            assert isinstance(section, ElasticSection)
            transf = GeomTransf(
                transf_type,
                self.model.uid_generator.new('transformation'),
                offset_i,
                offset_j,
                *axes)
            elm_el = ElasticBeamColumn(
                parent_component=assembly,
                uid=self.model.uid_generator.new('element'),
                nodes=[node_i, node_j],
                section=section,
                geomtransf=transf)
            res: Union[ElasticBeamColumn, DispBeamColumn] = elm_el
        elif element_type.__name__ == 'DispBeamColumn':
            assert isinstance(section, FiberSection)
            # TODO: add elastic section support
            transf = GeomTransf(
                transf_type,
                self.model.uid_generator.new('transformation'),
                offset_i,
                offset_j,
                *axes)
            beam_integration = Lobatto(
                uid=self.model.uid_generator.new('beam integration'),
                parent_section=section,
                n_p=5
            )
            elm_disp = DispBeamColumn(
                parent_component=assembly,
                uid=self.model.uid_generator.new('element'),
                nodes=[node_i, node_j],
                section=section,
                geomtransf=transf,
                integration=beam_integration)
            res = elm_disp
        return res

    def define_zerolength(
            self,
            assembly: ComponentAssembly,
            node_i: Node,
            node_j: Node,
            x_axis: nparr,
            y_axis: nparr,
            zerolength_gen: Callable,  # type: ignore
            zerolength_gen_args: dict[str, object]
    ):
        """
        Defines a zerolength element
        """
        dirs, mats = zerolength_gen(model=self.model, **zerolength_gen_args)
        elm = ZeroLength(
            assembly,
            self.model.uid_generator.new('element'),
            [node_i, node_j],
            mats,
            dirs,
            x_axis,
            y_axis
        )
        return elm

    def add_beamcolumn_elements_in_series(
            self,
            component,
            node_i,
            node_j,
            eo_i,
            eo_j,
            n_sub,
            transf_type,
            section,
            element_type,
            angle
    ):
        """
        Adds beamcolumn elemens in series
        """

        if n_sub > 1:
            p_i = np.array(node_i.coords) + eo_i
            p_j = np.array(node_j.coords) + eo_j
            internal_pt_coords = np.linspace(
                tuple(p_i),
                tuple(p_j),
                num=n_sub+1)
            intnodes = []
            for i in range(1, len(internal_pt_coords)-1):
                intnode = Node(self.model.uid_generator
                               .new('node'),
                               [*internal_pt_coords[i]])
                component.internal_nodes.add(intnode)
                intnodes.append(intnode)
        for i in range(n_sub):
            if i == 0:
                n_i = node_i
                o_i = eo_i
            else:
                n_i = intnodes[i-1]
                o_i = np.zeros(3)
            if i == n_sub - 1:
                n_j = node_j
                o_j = eo_j
            else:
                n_j = intnodes[i]
                o_j = np.zeros(3)
            element = self.define_beamcolumn(
                assembly=component,
                node_i=n_i, node_j=n_j,
                offset_i=o_i, offset_j=o_j, transf_type=transf_type,
                section=section,
                element_type=element_type,
                angle=angle)
            if element_type.__name__ == 'ElasticBeamColumn':
                component.elastic_beamcolumn_elements.add(element)
            elif element_type.__name__ == 'DispBeamColumn':
                component.disp_beamcolumn_elements.add(element)
            else:
                raise TypeError(
                    'Unsupported element type:'
                    f' {element_type.__name__}')

    def generate_plain_component_assembly(
            self,
            component_purpose,
            lvl,
            node_i,
            node_j,
            n_sub,
            eo_i,
            eo_j,
            section,
            element_type,
            transf_type,
            angle):
        """
        Generates a plain component assembly, with line elements in
        series
        """

        assert isinstance(node_i, Node)
        assert isinstance(node_j, Node)

        # instantiate a component assembly
        component = ComponentAssembly(
            uid=self.model.uid_generator.new('component'),
            parent_collection=lvl.components,
            component_purpose=component_purpose)
        # add it to the level
        lvl.components.add(component)
        # fill component assembly
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)

        self.add_beamcolumn_elements_in_series(
            component,
            node_i,
            node_j,
            eo_i,
            eo_j,
            n_sub,
            transf_type,
            section,
            element_type,
            angle
        )

    def generate_hinged_component_assembly(
            self,
            component_purpose,
            lvl,
            node_i,
            node_j,
            n_sub,
            eo_i,
            eo_j,
            section,
            element_type,
            transf_type,
            angle,
            zerolength_gen_i,
            zerolength_gen_args_i,
            zerolength_gen_j,
            zerolength_gen_args_j):
        """
        Defines a component assembly that is comprised of
        beamcolumn elements connected in series with nonlinear springs
        attached at the ends, followed by another sequence of
        beamcolumn elements (in order to be able to specify rigid offsets)
        """
        # instantiate a component assembly
        component = ComponentAssembly(
            uid=self.model.uid_generator.new('component'),
            parent_collection=lvl.components,
            component_purpose=component_purpose)
        # fill component assembly
        component.external_nodes.add(node_i)
        component.external_nodes.add(node_j)
        # add it to the level
        lvl.components.add(component)

        p_i = np.array(node_i.coords) + eo_i
        p_j = np.array(node_j.coords) + eo_j
        axes = local_axes_from_points_and_angle(
            p_i, p_j, angle)
        x_axis, y_axis, _ = axes
        clear_length = np.linalg.norm(p_j - p_i)
        zerolength_gen_args_i.update({'element_length': clear_length})
        zerolength_gen_args_j.update({'element_length': clear_length})

        # we can have hinges at both ends, or just one of the two ends.
        # ...or even no hinges!
        if zerolength_gen_i:
            hinge_location_i = p_i + x_axis * zerolength_gen_args_i['distance']
            nh_i_out = Node(self.model.uid_generator.new('node'),
                            [*hinge_location_i])
            nh_i_in = Node(self.model.uid_generator.new('node'),
                           [*hinge_location_i])
            nh_i_in.visibility.connected_to_zerolength = True
            component.internal_nodes.add(nh_i_out)
            component.internal_nodes.add(nh_i_in)
            self.add_beamcolumn_elements_in_series(
                component,
                node_i,
                nh_i_out,
                eo_i,
                np.zeros(3),
                zerolength_gen_args_i['n_sub'],
                transf_type,
                section,
                element_type,
                angle
            )
            zerolen_elm = self.define_zerolength(
                component,
                nh_i_out,
                nh_i_in,
                x_axis,
                y_axis,
                zerolength_gen_i,
                zerolength_gen_args_i
            )
            component.zerolength_elements.add(zerolen_elm)
            conn_node_i = nh_i_in
            conn_eo_i = np.zeros(3)
        else:
            conn_node_i = node_i
            conn_eo_i = eo_i
        if zerolength_gen_j:
            hinge_location_j = (p_i + x_axis *
                                (clear_length
                                 - zerolength_gen_args_j['distance']))
            nh_j_out = Node(self.model.uid_generator.new('node'),
                            [*hinge_location_j])
            nh_j_in = Node(self.model.uid_generator.new('node'),
                           [*hinge_location_j])
            nh_j_in.visibility.connected_to_zerolength = True
            component.internal_nodes.add(nh_j_out)
            component.internal_nodes.add(nh_j_in)
            self.add_beamcolumn_elements_in_series(
                component,
                nh_j_out,
                node_j,
                np.zeros(3),
                eo_j,
                zerolength_gen_args_j['n_sub'],
                transf_type,
                section,
                element_type,
                angle
            )
            zerolen_elm = self.define_zerolength(
                component,
                nh_j_out,
                nh_j_in,
                -x_axis,
                y_axis,
                zerolength_gen_j,
                zerolength_gen_args_j
            )
            component.zerolength_elements.add(zerolen_elm)
            conn_node_j = nh_j_in
            conn_eo_j = np.zeros(3)
        else:
            conn_node_j = node_j
            conn_eo_j = eo_j

        self.add_beamcolumn_elements_in_series(
            component,
            conn_node_i,
            conn_node_j,
            conn_eo_i,
            conn_eo_j,
            n_sub,
            transf_type,
            section,
            element_type,
            angle
        )

    def add_vertical_active(
            self,
            x_coord: float,
            y_coord: float,
            offset_i: nparr,
            offset_j: nparr,
            transf_type: str,
            n_sub: int,
            section: ElasticSection | FiberSection,
            element_type: Type[Union[ElasticBeamColumn, DispBeamColumn]],
            placement='centroid',
            angle=0.00,
            method='generate_plain_component_assembly',
            additional_args={}
    ):
        """
        Adds a vertical beamcolumn element to all active levels.  This
        method assumes that the levels are defined in order, from
        lowest to highest elevation, with consecutive ascending
        integer keys.
        """
        ndg = NodeGenerator(self.model)
        querry = ElmQuerry(self.model)
        lvls = self.model.levels
        assert lvls.active, 'No active levels.'
        for key in lvls.active:
            lvl = lvls[key]
            if key-1 not in lvls:
                continue

            top_node = querry.search_node_lvl(x_coord, y_coord, key)
            if not top_node:
                top_node = ndg.add_node_lvl(x_coord, y_coord, key)

            bottom_node = querry.search_node_lvl(x_coord, y_coord, key-1)
            if not bottom_node:
                bottom_node = ndg.add_node_lvl(x_coord, y_coord, key-1)

            # check for a panel zone
            top_node = look_for_panel_zone(top_node, lvl, querry)

            p_i = np.array(top_node.coords) + offset_i
            p_j = np.array(bottom_node.coords) + offset_j
            sec_offset_global = retrieve_snap_pt_global_offset(
                placement, section, p_i, p_j, angle)
            p_i += sec_offset_global
            p_j += sec_offset_global
            eo_i = offset_i + sec_offset_global
            eo_j = offset_j + sec_offset_global

            args = {
                'component_purpose': 'vertical_component',
                'lvl': lvl,
                'node_i': top_node,
                'node_j': bottom_node,
                'n_sub': n_sub,
                'eo_i': eo_i,
                'eo_j': eo_j,
                'section': section,
                'element_type': element_type,
                'transf_type': transf_type,
                'angle': angle
            }

            args.update(additional_args)
            assert hasattr(self, method), \
                f'Method not available: {method}'
            mthd = getattr(self, method)
            mthd(**args)

    def add_horizontal_active(
            self,
            xi_coord: float,
            yi_coord: float,
            xj_coord: float,
            yj_coord: float,
            offset_i: nparr,
            offset_j: nparr,
            snap_i: str,
            snap_j: str,
            transf_type: str,
            n_sub: int,
            section: ElasticSection,
            element_type: Type[Union[ElasticBeamColumn, DispBeamColumn]],
            placement='centroid',
            angle=0.00,
            split_existing_i=None,
            split_existing_j=None,
            method='generate_plain_component_assembly',
            additional_args={}
    ):
        """
        Adds a horizontal beamcolumn element to all active levels.
        """
        querry = ElmQuerry(self.model)
        ndg = NodeGenerator(self.model)
        lvls = self.model.levels
        assert lvls.active, 'No active levels.'
        for key in lvls.active:
            lvl = lvls[key]

            p_i_init = np.array((xi_coord, yi_coord, lvl.elevation)) + offset_i
            p_j_init = np.array((xj_coord, yj_coord, lvl.elevation)) + offset_j

            if section.snap_points and (placement != 'centroid'):
                # obtain offset from section (local system)
                d_z, d_y = section.snap_points[placement]
                sec_offset_local: nparr = np.array([0.00, d_y, d_z])
                # retrieve local coordinate system
                x_axis, y_axis, z_axis = \
                    local_axes_from_points_and_angle(
                        p_i_init, p_j_init, angle)  # type: ignore
                t_glob_to_loc = transformation_matrix(
                    x_axis, y_axis, z_axis)
                t_loc_to_glob = t_glob_to_loc.T
                sec_offset_global = t_loc_to_glob @ sec_offset_local
            else:
                sec_offset_global = np.zeros(3)

            node_i, eo_i = beam_placement_lookup(
                xi_coord, yi_coord, querry, ndg,
                lvls, key, offset_i,
                sec_offset_global,
                split_existing_i,
                snap_i)
            node_j, eo_j = beam_placement_lookup(
                xj_coord, yj_coord, querry, ndg,
                lvls, key, offset_j,
                sec_offset_global,
                split_existing_j,
                snap_j)

            args = {
                'component_purpose': 'horizontal_component',
                'lvl': lvl,
                'node_i': node_i,
                'node_j': node_j,
                'n_sub': n_sub,
                'eo_i': eo_i,
                'eo_j': eo_j,
                'section': section,
                'element_type': element_type,
                'transf_type': transf_type,
                'angle': angle
            }

            args.update(additional_args)
            assert hasattr(self, method), \
                f'Method not available: {method}'
            mthd = getattr(self, method)
            mthd(**args)

    def add_pz_active(
            self,
            x_coord: float,
            y_coord: float,
            section: ElasticSection,
            physical_material,
            angle,
            column_depth,
            beam_depth,
            pz_doubler_plate_thickness: float,
            pz_hardening: float
    ):
        """
        Adds a component assembly representing a steel W-section
        panel zone joint.
        """
        ndg = NodeGenerator(self.model)
        querry = ElmQuerry(self.model)
        lvls = self.model.levels
        assert lvls.active, 'No active levels.'
        for key in lvls.active:

            lvl = lvls[key]
            if key-1 not in lvls:
                continue

            top_node = querry.search_node_lvl(x_coord, y_coord, key)
            if not top_node:
                top_node = ndg.add_node_lvl(x_coord, y_coord, key)

            # instantiate a component assembly
            component = ComponentAssembly(
                uid=self.model.uid_generator.new('component'),
                parent_collection=lvl.components,
                component_purpose='steel_W_panel_zone')
            # add it to the level
            lvl.components.add(component)

            p_i: nparr = np.array(top_node.coords)
            p_j = (np.array(top_node.coords)
                   + np.array((0.00, 0.00, -beam_depth)))
            x_axis, y_axis, z_axis = \
                local_axes_from_points_and_angle(
                    p_i, p_j, angle)  # type: ignore

            # determine node locations
            top_h_f_loc = p_i + y_axis * column_depth/2.00
            top_h_b_loc = p_i - y_axis * column_depth/2.00
            top_v_f_loc = p_i + y_axis * column_depth/2.00
            top_v_b_loc = p_i - y_axis * column_depth/2.00
            mid_v_f_loc = (p_i + y_axis * column_depth/2.00
                           + x_axis * beam_depth/2.00)
            mid_v_b_loc = (p_i - y_axis * column_depth/2.00
                           + x_axis * beam_depth/2.00)
            bottom_h_f_loc = (p_i + y_axis * column_depth/2.00
                              + x_axis * beam_depth)
            bottom_h_b_loc = (p_i - y_axis * column_depth/2.00
                              + x_axis * beam_depth)
            bottom_v_f_loc = (p_i + y_axis * column_depth/2.00
                              + x_axis * beam_depth)
            bottom_v_b_loc = (p_i - y_axis * column_depth/2.00
                              + x_axis * beam_depth)

            # define nodes
            top_h_f = Node(
                self.model.uid_generator.new('node'),
                [*top_h_f_loc]
            )
            top_h_b = Node(
                self.model.uid_generator.new('node'),
                [*top_h_b_loc]
            )
            top_v_f = Node(
                self.model.uid_generator.new('node'),
                [*top_v_f_loc]
            )
            top_v_f.visibility.connected_to_zerolength = True
            top_v_b = Node(
                self.model.uid_generator.new('node'),
                [*top_v_b_loc]
            )
            top_v_b.visibility.connected_to_zerolength = True

            mid_v_f = ndg.add_node_lvl_xyz(
                mid_v_f_loc[0],
                mid_v_f_loc[1],
                mid_v_f_loc[2],
                lvl.uid)
            mid_v_b = ndg.add_node_lvl_xyz(
                mid_v_b_loc[0],
                mid_v_b_loc[1],
                mid_v_b_loc[2],
                lvl.uid)

            bottom_h_f = Node(
                self.model.uid_generator.new('node'),
                [*bottom_h_f_loc]
            )
            bottom_h_b = Node(
                self.model.uid_generator.new('node'),
                [*bottom_h_b_loc]
            )
            bottom_v_f = Node(
                self.model.uid_generator.new('node'),
                [*bottom_v_f_loc]
            )
            bottom_v_f.visibility.connected_to_zerolength = True
            bottom_v_b = Node(
                self.model.uid_generator.new('node'),
                [*bottom_v_b_loc]
            )
            bottom_v_b.visibility.connected_to_zerolength = True

            bottom_mid = ndg.add_node_lvl_xyz(
                p_j[0], p_j[1], p_j[2], lvl.uid)

            # define rigid beamcolumn elements
            if not self.model.elastic_sections.retrieve_by_attr(
                    'name', 'rigid_link_section'):
                load_util_rigid_elastic(self.model)
            rigid_sec = self.model.elastic_sections.retrieve_by_attr(
                'name', 'rigid_link_section')
            assert rigid_sec

            elm_top_h_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [top_node, top_h_f],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    y_axis, -x_axis, z_axis
                )
            )
            elm_top_h_f.visibility.hidden_when_extruded = True

            elm_top_h_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [top_h_b, top_node],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    y_axis, -x_axis, z_axis
                )
            )
            elm_top_h_b.visibility.hidden_when_extruded = True

            elm_bottom_h_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [bottom_mid, bottom_h_f],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    y_axis, -x_axis, z_axis
                )
            )
            elm_bottom_h_f.visibility.hidden_when_extruded = True

            elm_bottom_h_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [bottom_h_b, bottom_mid],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    y_axis, -x_axis, z_axis
                )
            )
            elm_bottom_h_b.visibility.hidden_when_extruded = True

            elm_top_v_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [top_v_f, mid_v_f],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    x_axis, y_axis, z_axis
                )
            )
            elm_top_v_f.visibility.hidden_when_extruded = True

            elm_top_v_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [top_v_b, mid_v_b],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    x_axis, y_axis, z_axis
                )
            )
            elm_top_v_b.visibility.hidden_when_extruded = True

            elm_bottom_v_f = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [mid_v_f, bottom_v_f],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    x_axis, y_axis, z_axis
                )
            )
            elm_bottom_v_f.visibility.hidden_when_extruded = True

            elm_bottom_v_b = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [mid_v_b, bottom_v_b],
                rigid_sec,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    x_axis, y_axis, z_axis
                )
            )
            elm_bottom_v_b.visibility.hidden_when_extruded = True

            elm_interior = ElasticBeamColumn(
                component,
                self.model.uid_generator.new('element'),
                [top_node, bottom_mid],
                section,
                GeomTransf(
                    'Corotational',
                    self.model.uid_generator.new('transformation'),
                    np.zeros(3), np.zeros(3),
                    x_axis, y_axis, z_axis
                )
            )
            elm_interior.visibility.skip_opensees_definition = True
            elm_interior.visibility.hidden_at_line_plots = True

            # define zerolength elements
            zerolen_top_f = self.define_zerolength(
                component,
                top_h_f,
                top_v_f,
                x_axis,
                y_axis,
                steel_w_col_pz,
                {
                    'section': section,
                    'physical_material': physical_material,
                    'pz_length': beam_depth,
                    'pz_doubler_plate_thickness': pz_doubler_plate_thickness,
                    'pz_hardening': pz_hardening
                 }
            )
            zerolen_top_b = self.define_zerolength(
                component,
                top_h_b,
                top_v_b,
                x_axis,
                y_axis,
                steel_w_col_pz,
                {
                    'section': section,
                    'physical_material': physical_material,
                    'pz_length': beam_depth,
                    'pz_doubler_plate_thickness': pz_doubler_plate_thickness,
                    'pz_hardening': pz_hardening
                 }
            )
            zerolen_bottom_f = self.define_zerolength(
                component,
                bottom_h_f,
                bottom_v_f,
                x_axis,
                y_axis,
                steel_w_col_pz,
                {
                    'section': section,
                    'physical_material': physical_material,
                    'pz_length': beam_depth,
                    'pz_doubler_plate_thickness': pz_doubler_plate_thickness,
                    'pz_hardening': pz_hardening
                 }
            )
            zerolen_bottom_b = self.define_zerolength(
                component,
                bottom_h_b,
                bottom_v_b,
                x_axis,
                y_axis,
                steel_w_col_pz,
                {
                    'section': section,
                    'physical_material': physical_material,
                    'pz_length': beam_depth,
                    'pz_doubler_plate_thickness': pz_doubler_plate_thickness,
                    'pz_hardening': pz_hardening
                 }
            )

            # fill component assembly
            component.external_nodes.add(top_node)
            component.external_nodes.named_contents['top_node'] = top_node
            component.external_nodes.add(bottom_mid)
            component.external_nodes.named_contents['bottom_node'] = bottom_mid
            component.external_nodes.add(mid_v_f)
            component.external_nodes.named_contents['middle_front'] = mid_v_f
            component.external_nodes.add(mid_v_b)
            component.external_nodes.named_contents['middle_back'] = mid_v_b

            component.internal_nodes.add(top_h_f)
            component.internal_nodes.add(top_h_b)
            component.internal_nodes.add(top_v_f)
            component.internal_nodes.add(top_v_b)
            component.internal_nodes.add(bottom_h_f)
            component.internal_nodes.add(bottom_h_b)
            component.internal_nodes.add(bottom_v_f)
            component.internal_nodes.add(bottom_v_b)

            component.elastic_beamcolumn_elements.add(elm_top_h_f)
            (component.elastic_beamcolumn_elements
             .named_contents['elm_top_h_f']) = elm_top_h_f
            component.elastic_beamcolumn_elements.add(elm_top_h_b)
            (component.elastic_beamcolumn_elements
             .named_contents['elm_top_h_b']) = elm_top_h_b
            component.elastic_beamcolumn_elements.add(elm_bottom_h_f)
            component.elastic_beamcolumn_elements.add(elm_bottom_h_b)
            component.elastic_beamcolumn_elements.add(elm_top_v_f)
            component.elastic_beamcolumn_elements.add(elm_top_v_b)
            component.elastic_beamcolumn_elements.add(elm_bottom_v_f)
            component.elastic_beamcolumn_elements.add(elm_bottom_v_b)
            component.elastic_beamcolumn_elements.add(elm_interior)
            (component.elastic_beamcolumn_elements
             .named_contents['elm_interior']) = elm_interior

            component.zerolength_elements.add(zerolen_top_f)
            (component.zerolength_elements
             .named_contents['nonlinear_spring']) = zerolen_top_f
            component.zerolength_elements.add(zerolen_top_b)
            component.zerolength_elements.add(zerolen_bottom_f)
            component.zerolength_elements.add(zerolen_bottom_b)



"""
Model Generator for OpenSees ~ node generator
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass
from ..ops.node import Node
if TYPE_CHECKING:
    from .model import Model


@dataclass(repr=False)
class NodeGenerator:
    """
    Introduces node objects to a model.
    """
    model: Model

    def add_node_lvl(self, x_loc: float, y_loc: float, lvl: int) -> Node:
        """
        Adds a node at the specified coordinates and level
        """
        lvls = self.model.levels
        level = lvls[lvl]
        node = Node(
            uid=self.model.uid_generator.new('node'),
            coords=[x_loc, y_loc, level.elevation])
        level.nodes.add(node)
        return node

    def add_node_lvl_xyz(self, x_loc: float, y_loc: float,
                         z_loc: float, lvl: int) -> Node:
        """
        Adds a node at the specified coordinates and level
        having a custo elevation
        """
        lvls = self.model.levels
        level = lvls[lvl]
        node = Node(
            uid=self.model.uid_generator.new('node'),
            coords=[x_loc, y_loc, z_loc])
        level.nodes.add(node)
        return node

    def add_node_active(self, x_loc, y_loc):
        """
        Adds a node at the specified coordinates to all active levels.
        """
        lvls = self.model.levels
        assert lvls.active, 'No active levels.'
        for key in lvls.active:
            self.add_node_lvl(x_loc, y_loc, key)



"""
Model Generator for OpenSees ~ node
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass, field
from functools import total_ordering
from ..graphics.visibility import NodeVisibility


@dataclass
@total_ordering
class Node:
    """
    OpenSees node
    https://openseespydoc.readthedocs.io/en/latest/src/node.html?highlight=node
    Attributes:
        uid (int)
        coords (list[float])
        mass (list[float])
        load (list[float])
        restraint (list[bool])
    """
    uid: int
    coords: list[float]
    restraint: list[bool] = field(init=False)
    visibility: NodeVisibility = field(default_factory=NodeVisibility)

    def __post_init__(self):
        self.restraint = [False]*6
        self.mass = [0.00]*6

    def __le__(self, other):
        return self.uid < other.uid

    def __repr__(self):
        res = ''
        res += 'Node object\n'
        res += f'  uid: {self.uid}\n'
        res += f'  coords: {self.coords}\n'
        res += f'  restraint: {self.restraint}\n'
        return res



"""
Toy structures
Predefined models that are used to demonstrate things in the example
notebooks.  They are used to avoid littering the example notebooks
with code that just defines a model, so that they can focus on the
important things.
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

import sys
sys.path.append('../src')
import model
import numpy as np


# Simple 3-story 4-bay moment frame
smrf_3_4 = model.Model()
smrf_3_4.add_level('base', 0.00, 'fixed')
smrf_3_4.add_level('1', 144.00)
smrf_3_4.add_level('2', 144.00*2)
smrf_3_4.add_level('3', 144.00*3)

smrf_3_4.set_active_material('steel-bilinear-fy50')
smrf_3_4.add_sections_from_json(
    "../section_data/sections.json",
    'W',
    ["W24X94"])
smrf_3_4.set_active_section("W24X94")
smrf_3_4.set_active_levels("all_above_base")
smrf_3_4.active_placement = 'centroid'
smrf_3_4.set_active_section("W24X94")
p1 = np.array((0.00, 0.00))
p2 = np.array((360., 0.00))
p3 = np.array((360.*2., 0.00))
p4 = np.array((360.*3, 0.00))
smrf_3_4.active_angle = np.pi/2
for pt in [p1, p2, p3, p4]:
    smrf_3_4.add_column_at_point(
        pt,
        n_sub=4,
        geom_transf='Corotational',
        ends={'type': 'fixed', 'doubler plate thickness': 0.00})
smrf_3_4.active_placement = 'top_center'

smrf_3_4.active_angle = 0.00
for pair in ((p1, p2), (p2, p3), (p3, p4)):
    smrf_3_4.add_beam_at_points(
        pair[0], pair[1],
        n_sub=4,
        snap_i='bottom_center',
        snap_j='top_center')



"""
Model Generator for OpenSees ~ defaults
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
import numpy as np
import numpy.typing as npt
from .physical_material import PhysicalMaterial
from . import common
from .ops.uniaxial_material import Elastic
from .ops.uniaxial_material import Steel02
from .ops.section import ElasticSection
from .gen.mesh_shapes import rect_mesh
if TYPE_CHECKING:
    from .model import Model

nparr = npt.NDArray[np.float64]


def load_util_rigid_elastic(model: Model):
    """
    Adds a default rigid elastic beamcolumn element
    to the model
    """
    new_uid = model.uid_generator.new('section')
    sec = ElasticSection(
        name='rigid_link_section',
        uid=new_uid,
        outside_shape=None,
        snap_points=None,
        e_mod=common.STIFF,
        area=1.00,
        i_y=1.00,
        i_x=1.00,
        g_mod=common.STIFF,
        j_mod=1.00,
        sec_w=0.00)
    model.elastic_sections.add(sec)


def load_default_elastic(model: Model, sec_name: str):
    """
    Adds default non-rigid elastic beamcolumn element
    sections to the model
    """
    new_uid = model.uid_generator.new('section')
    sec = ElasticSection(
        name=sec_name,
        uid=new_uid,
        outside_shape=None,
        snap_points=None,
        e_mod=1.00,
        area=1.00,
        i_y=1.00,
        i_x=1.00,
        g_mod=1.00,
        j_mod=1.00,
        sec_w=0.00)
    if model.settings.imperial_units:
        y_max = +10.00
        y_min = -10.00
        z_max = +6.00
        z_min = -6.00
        sec.outside_shape = rect_mesh(12.0, 20.0)
    else:
        y_max = +0.25
        y_min = -0.25
        z_max = +0.15
        z_min = -0.15
        sec.outside_shape = rect_mesh(0.30, 0.50)
    snap_points: dict[str, nparr] = {
        'centroid': np.array([0., 0.]),
        'top_center': np.array([0., -y_max]),
        'top_left': np.array([-z_min, -y_max]),
        'top_right': np.array([-z_max, -y_max]),
        'center_left': np.array([-z_min, 0.]),
        'center_right': np.array([-z_max, 0.]),
        'bottom_center': np.array([0., -y_min]),
        'bottom_left': np.array([-z_min, -y_min]),
        'bottom_right': np.array([-z_max, -y_min])
    }
    sec.snap_points = snap_points
    model.elastic_sections.add(sec)


def load_default_steel(model: Model):
    """
    Adds a default steel material to the model
    """
    # TODO: add SI units some day
    uniaxial_mat = Steel02(
        uid=model.uid_generator.new('uniaxial material'),
        name='default steel',
        Fy=55000.00,
        E0=29000000.00,
        b=0.01,
        params=(19.0, 0.925, 0.15),
        a1=0.12,
        a2=0.90,
        a3=0.18,
        a4=0.90,
        sigInit=0.0,
        G=11153846.15)
    physical_mat = PhysicalMaterial(
        model.uid_generator.new('physical material'),
        'default steel',
        'A992-Fy50',
        0.2835648148148148/common.G_CONST_IMPERIAL,
        29000000.00,
        11153846.15,
        55000.00)
    model.uniaxial_materials.add(uniaxial_mat)
    model.physical_materials.add(physical_mat)


def load_default_fix_release(model: Model):
    """
    Loads default fix and release elastic uniaxial materials
    used to simulate moment releases using zerolength elements.
    """
    uniaxial_mat = Elastic(
        uid=model.uid_generator.new('uniaxial material'),
        name='fix',
        e_mod=common.STIFF_ROT)
    model.uniaxial_materials.add(uniaxial_mat)
    uniaxial_mat = Elastic(
        uid=model.uid_generator.new('uniaxial material'),
        name='release',
        e_mod=common.TINY)
    model.uniaxial_materials.add(uniaxial_mat)



"""
Model Generator for OpenSees ~ uid generator
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass
from itertools import count


@dataclass
class UIDGenerator:
    """
    Generates unique identifiers, uids, for various things.
    """

    def new(self, thing: str):
        """
        Provide a uid for a new node
        """
        if hasattr(self, thing):
            res = next(getattr(self, thing))
        else:
            setattr(self, thing, count(0))
            res = next(getattr(self, thing))
        return res



"""
Model Generator for OpenSees ~ element
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from dataclasses import dataclass, field
from typing import Union
import numpy as np
import numpy.typing as npt
from .uniaxial_material import UniaxialMaterial
from .node import Node
from .section import ElasticSection
from .section import FiberSection
from ..graphics.visibility import ElementVisibility
from .. import component_assembly


nparr = npt.NDArray[np.float64]


@dataclass(repr=False)
class Element:
    """
    OpenSees element
    https://openseespydoc.readthedocs.io/en/latest/src/element.html
    """
    parent_component: component_assembly.ComponentAssembly = field(repr=False)
    uid: int
    nodes: list[Node]


@dataclass(repr=False)
class ZeroLength(Element):
    """
    OpenSees ZeroLength element
    https://openseespydoc.readthedocs.io/en/latest/src/ZeroLength.html
    """
    mats: list[UniaxialMaterial]
    dirs: list[int]
    vecx: nparr
    vecyp: nparr

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'zeroLength',
            self.uid,
            *[n.uid for n in self.nodes],
            '-mat',
            *[m.uid for m in self.mats],
            '-dir',
            *self.dirs,
            '-orient',
            *self.vecx,
            *self.vecyp
        ]

    def __repr__(self):
        res = ''
        res += 'ZeroLength element object\n'
        res += f'uid: {self.uid}'
        res += 'Materials:'
        for mat, direction in zip(self.mats, self.dirs):
            res += f'  {direction}: {mat.name}\n'
        res += f'vecx: {self.vecx}\n'
        res += f'vecyp: {self.vecyp}\n'
        return res


@dataclass(repr=False)
class GeomTransf:
    """
    OpenSees geomTransf object
    https://openseespydoc.readthedocs.io/en/latest/src/ZeroLength.html
    https://openseespydoc.readthedocs.io/en/latest/src/geomTransf.html?highlight=geometric%20transformation
    """
    transf_type: str
    uid: int
    offset_i: nparr
    offset_j: nparr
    x_axis: nparr
    y_axis: nparr
    z_axis: nparr

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            self.transf_type,
            self.uid,
            *self.z_axis,
            '-jntOffset',
            *self.offset_i,
            *self.offset_j
        ]


@dataclass(repr=False)
class ElasticBeamColumn(Element):
    """
    OpenSees Elastic Beam Column Element
    https://openseespydoc.readthedocs.io/en/latest/src/elasticBeamColumn.html
    """
    section: ElasticSection
    geomtransf: GeomTransf
    visibility: ElementVisibility = field(
        default_factory=ElementVisibility)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'elasticBeamColumn',
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.section.area,
            self.section.e_mod,
            self.section.g_mod,
            self.section.j_mod,
            self.section.i_y,
            self.section.i_x,
            self.geomtransf.uid
        ]

    def clear_length(self):
        """
        Returns the clear length of the element (without the rigid
        offsets)
        """
        p_i = np.array(self.nodes[0].coords) + self.geomtransf.offset_i
        p_j = np.array(self.nodes[1].coords) + self.geomtransf.offset_j
        return np.linalg.norm(p_i - p_j)

    def __repr__(self):
        res = ''
        res += 'elasticBeamColumn element object\n'
        res += f'uid: {self.uid}\n'
        res += f'node_i.uid: {self.nodes[0].uid}\n'
        res += f'node_j.uid: {self.nodes[1].uid}\n'
        res += f'node_i.coords: {self.nodes[0].coords}\n'
        res += f'node_j.coords: {self.nodes[1].coords}\n'
        res += f'offset_i: {self.geomtransf.offset_i}\n'
        res += f'offset_j: {self.geomtransf.offset_j}\n'
        res += f'x_axis: {self.geomtransf.x_axis}\n'
        res += f'y_axis: {self.geomtransf.y_axis}\n'
        res += f'z_axis: {self.geomtransf.z_axis}\n'
        res += f'section.name: {self.section.name}\n'
        return res


@dataclass
class BeamIntegration:
    """
    OpenSees beamIntegration parent class.
    https://openseespydoc.readthedocs.io/en/latest/src/beamIntegration.html?highlight=beamintegration
    """
    uid: int
    parent_section: Union[ElasticSection, FiberSection] = field(repr=False)


@dataclass
class Lobatto(BeamIntegration):
    """
    OpenSees Lobatto beam integration.
    https://openseespydoc.readthedocs.io/en/latest/src/Lobatto.html
    """
    n_p: int

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'Lobatto',
            self.uid,
            self.parent_section.uid,
            self.n_p
        ]


@dataclass
class DispBeamColumn(Element):
    """
    OpenSees dispBeamColumn element
    https://openseespydoc.readthedocs.io/en/latest/src/ForceBeamColumn.html
    """
    section: FiberSection
    geomtransf: GeomTransf
    integration: BeamIntegration
    visibility: ElementVisibility = field(
        default_factory=ElementVisibility)

    def ops_args(self):
        """
        Returns the arguments required to define the object in
        OpenSees
        """
        return [
            'dispBeamColumn',
            self.uid,
            self.nodes[0].uid,
            self.nodes[1].uid,
            self.geomtransf.uid,
            self.integration.uid,
        ]

    def clear_length(self):
        """
        Returns the clear length of the element (without the rigid
        offsets)
        """
        p_i = np.array(self.nodes[0].coords) + self.geomtransf.offset_i
        p_j = np.array(self.nodes[1].coords) + self.geomtransf.offset_j
        return np.linalg.norm(p_i - p_j)

    def __repr__(self):
        res = ''
        res += 'dispBeamColumn element object\n'
        res += f'uid: {self.uid}\n'
        res += f'node_i.uid: {self.nodes[0].uid}\n'
        res += f'node_j.uid: {self.nodes[1].uid}\n'
        res += f'node_i.coords: {self.nodes[0].coords}\n'
        res += f'node_j.coords: {self.nodes[1].coords}\n'
        res += f'offset_i: {self.geomtransf.offset_i}\n'
        res += f'offset_j: {self.geomtransf.offset_j}\n'
        res += f'x_axis: {self.geomtransf.x_axis}\n'
        res += f'y_axis: {self.geomtransf.y_axis}\n'
        res += f'z_axis: {self.geomtransf.z_axis}\n'
        res += f'section.name: {self.section.name}\n'
        return res



"""
The following utility functions are used for data visualization
https://plotly.com/python/reference/
"""
#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from typing import Optional
import sys
import plotly.graph_objects as go  # type: ignore
import numpy as np
import numpy.typing as npt
from . import graphics_common, graphics_common_3d
if TYPE_CHECKING:
    from model import Model
    from ..load_case import LoadCase

nparr = npt.NDArray[np.float64]


def add_data__nodes(data_dict, mdl, load_case):
    """
    Adds a trace containing nodes
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
      load_case (LoadCase): the load_case to be visualized
    """
    # TODO: this was written very sloppily, because I was in a rush.
    # I need to improve this code, avoid code repetition
    # TODO: merge this method with the other methods that plot nodes.
    list_of_nodes = mdl.list_of_primary_nodes()
    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    customdata_lst = []
    restraints = [node.restraint for node in list_of_nodes]
    restraint_symbols = []
    for node in list_of_nodes:
        if True in node.restraint:
            restraint_symbols.append("fixed")
        else:
            restraint_symbols.append("free")
    for node in list_of_nodes:
        if load_case:
            customdata_lst.append(
                (node.uid,
                 *load_case.node_mass[node.uid].val,
                 *load_case.node_loads[node.uid].val
                 )
            )
        else:
            customdata_lst.append(
                (node.uid,
                 )
            )
    customdata: nparr = np.array(customdata_lst, dtype='object')
    if load_case:
        data_dict.append({
            "name": "Primary nodes",
            "type": "scatter3d",
            "mode": "markers",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "customdata": customdata,
            "text": restraints,
            "hovertemplate": 'Coordinates: '
            '(%{x:.2f}, %{y:.2f}, %{z:.2f})<br>' +
            'Restraint: %{text}<br>' +
            'Mass: (%{customdata[1]:.3g}, ' +
            '%{customdata[2]:.3g}, %{customdata[3]:.3g}, ' +
            '%{customdata[4]:.3g}, %{customdata[5]:.3g}, ' +
            '%{customdata[6]:.3g})<br>' +
            'Load: (%{customdata[7]:.3g}, ' +
            '%{customdata[8]:.3g}, %{customdata[9]:.3g}, ' +
            '%{customdata[10]:.3g}, %{customdata[11]:.3g}, ' +
            '%{customdata[12]:.3g})' +
            '<extra>Node: %{customdata[0]:d}</extra>',
            "marker": {
                "symbol": [graphics_common_3d.node_marker[sym][0]
                           for sym in restraint_symbols],
                "color": graphics_common.NODE_PRIMARY_COLOR,
                "size": [graphics_common_3d.node_marker[sym][1]
                         for sym in restraint_symbols],
                "line": {
                    "color": graphics_common.NODE_PRIMARY_COLOR,
                    "width": 4}
            }
        })
    else:
        data_dict.append({
            "name": "Primary nodes",
            "type": "scatter3d",
            "mode": "markers",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "customdata": customdata,
            "text": restraints,
            "hovertemplate": 'Coordinates: '
            '(%{x:.2f}, %{y:.2f}, %{z:.2f})<br>' +
            'Restraint: %{text}<br>' +
            '<extra>Node: %{customdata[0]:d}</extra>',
            "marker": {
                "symbol": [graphics_common_3d.node_marker[sym][0]
                           for sym in restraint_symbols],
                "color": graphics_common.NODE_PRIMARY_COLOR,
                "size": [graphics_common_3d.node_marker[sym][1]
                         for sym in restraint_symbols],
                "line": {
                    "color": graphics_common.NODE_PRIMARY_COLOR,
                    "width": 4}
            }
        })


def add_data__parent_nodes(data_dict, load_case: LoadCase):
    """
    Adds a trace containing parent nodes
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
      load_case (LoadCase): the load_case to be visualized
    """
    list_of_nodes = load_case.parent_nodes.values()
    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    customdata_list = []
    restraints = [node.restraint for node in list_of_nodes]
    for node in list_of_nodes:
        customdata_list.append(
            (node.uid,
             *load_case.node_mass[node.uid].val,
             *load_case.node_loads[node.uid].val
             )
        )

    customdata: nparr = np.array(customdata_list, dtype='object')
    data_dict.append({
        "type": "scatter3d",
        "mode": "markers",
        "x": x_list,
        "y": y_list,
        "z": z_list,
        "customdata": customdata,
        "text": restraints,
        "hovertemplate": 'Coordinates: (%{x:.2f}, %{y:.2f}, %{z:.2f})<br>' +
        'Restraint: %{text}<br>' +
        'Mass: (%{customdata[1]:.3g}, ' +
        '%{customdata[2]:.3g}, %{customdata[3]:.3g}, ' +
        '%{customdata[4]:.3g}, %{customdata[5]:.3g}, ' +
        '%{customdata[6]:.3g})<br>' +
        'Load: (%{customdata[7]:.3g}, ' +
        '%{customdata[8]:.3g}, %{customdata[9]:.3g}, ' +
        '%{customdata[10]:.3g}, %{customdata[11]:.3g}, ' +
        '%{customdata[12]:.3g})' +
        '<extra>Parent Node: %{customdata[0]:d}</extra>',
        "marker": {
            "symbol": [graphics_common_3d.node_marker['parent'][0]
                       for node in list_of_nodes],
            "color": graphics_common.NODE_PRIMARY_COLOR,
            "size": [graphics_common_3d.node_marker['parent'][1]
                     for node in list_of_nodes],
            "line": {
                "color": graphics_common.NODE_PRIMARY_COLOR,
                "width": 4}
        }
    })


def add_data__internal_nodes(data_dict, mdl, load_case):
    """
    Adds a trace containing internal nodes
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
      load_case (LoadCase): the load_case to be visualized
    """
    list_of_nodes = mdl.list_of_internal_nodes()
    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    customdata = []
    restraints = [node.restraint for node in list_of_nodes]
    restraint_symbols = []
    for node in list_of_nodes:
        if True in node.restraint:
            restraint_symbols.append("fixed")
        elif node.visibility.connected_to_zerolength:
            restraint_symbols.append("release")
        else:
            restraint_symbols.append("internal")
    for node in list_of_nodes:
        if load_case:
            customdata.append(
                (node.uid,
                 *load_case.node_mass[node.uid].val,
                 *load_case.node_loads[node.uid].val
                 )
            )
        else:
            customdata.append(
                (node.uid,
                 )
            )
    if load_case:
        data_dict.append({
            "name": "Internal nodes",
            "type": "scatter3d",
            "mode": "markers",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "customdata": customdata,
            "text": restraints,
            "hovertemplate": 'Coordinates: '
            '(%{x:.2f}, %{y:.2f}, %{z:.2f})<br>' +
            'Restraint: %{text}<br>' +
            'Mass: (%{customdata[1]:.3g}, ' +
            '%{customdata[2]:.3g}, %{customdata[3]:.3g})<br>' +
            'Load: (%{customdata[4]:.3g}, ' +
            '%{customdata[5]:.3g}, %{customdata[6]:.3g})' +
            '<extra>Node: %{customdata[0]:d}</extra>',
            "marker": {
                "symbol": [graphics_common_3d.node_marker[sym][0]
                           for sym in restraint_symbols],
                "color": graphics_common.NODE_INTERNAL_COLOR,
                "size": [graphics_common_3d.node_marker[sym][1]
                         for sym in restraint_symbols],
                "line": {
                    "color": graphics_common.NODE_INTERNAL_COLOR,
                    "width": 2}
            }
        })
    else:
        data_dict.append({
            "name": "Internal nodes",
            "type": "scatter3d",
            "mode": "markers",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "customdata": customdata,
            "text": restraints,
            "hovertemplate": 'Coordinates: '
            '(%{x:.2f}, %{y:.2f}, %{z:.2f})<br>' +
            'Restraint: %{text}<br>' +
            '<extra>Node: %{customdata[0]:d}</extra>',
            "marker": {
                "symbol": [graphics_common_3d.node_marker[sym][0]
                           for sym in restraint_symbols],
                "color": graphics_common.NODE_INTERNAL_COLOR,
                "size": [graphics_common_3d.node_marker[sym][1]
                         for sym in restraint_symbols],
                "line": {
                    "color": graphics_common.NODE_INTERNAL_COLOR,
                    "width": 2}
            }
        })


def add_data__release_nodes(data_dict, list_of_nodes):
    """
    Adds a trace containing release nodes
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
      load_case (LoadCase): the load_case to be visualized
    """
    x_list = [node.coords[0] for node in list_of_nodes]
    y_list = [node.coords[1] for node in list_of_nodes]
    z_list = [node.coords[2] for node in list_of_nodes]
    customdata = []
    for node in list_of_nodes:
        customdata.append(
            (node.uid,
             *node.mass,
             *node.load_total()
             )
        )
    data_dict.append({
        "type": "scatter3d",
        "mode": "markers",
        "x": x_list,
        "y": y_list,
        "z": z_list,
        "hoverinfo": "skip",
        "marker": {
            "symbol": graphics_common_3d.node_marker['pinned'][0],
            "color": graphics_common.NODE_INTERNAL_COLOR,
            "size": graphics_common_3d.node_marker['pinned'][1],
            "line": {
                "color": graphics_common.NODE_INTERNAL_COLOR,
                "width": 2}
        }
    })


# def add_data__diaphragm_lines(data_dict, lvl):
#     """
#     Adds a trace containing rigid diaphragm indicator lines
#     Arguments:
#       data_dict (dict): dictionary containing figure data
#       lvl (Level): a level
#     """
#     if not lvl.parent_node:
#         return
#     parent_node = lvl.parent_node
#     x_list: list[Optional[float]] = []
#     y_list: list[Optional[float]] = []
#     z_list: list[Optional[float]] = []
#     for node in lvl.nodes_primary.values():
#         x_list.extend(
#             (node.coords[0], parent_node.coords[0], None)
#         )
#         y_list.extend(
#             (node.coords[1], parent_node.coords[1], None)
#         )
#         z_list.extend(
#             (node.coords[2], parent_node.coords[2], None)
#         )
#     data_dict.append({
#         "type": "scatter3d",
#         "mode": "lines",
#         "x": x_list,
#         "y": y_list,
#         "z": z_list,
#         "hoverinfo": "skip",
#         "line": {
#             "width": 4,
#             "dash": "dash",
#             "color": graphics_common.GRID_COLOR
#         }
#     })


# def add_data__bisector_lines(data_dict, lvl):
#     """
#     Adds a trace containing tributary area bisector lines
#     Arguments:
#       data_dict (dict): dictionary containing figure data
#       lvl (Level): a level
#     """
#     if not lvl.parent_node:
#         return
#     x_list: list[Optional[float]] = []
#     y_list: list[Optional[float]] = []
#     z_list: list[Optional[float]] = []
#     for line in lvl.floor_bisector_lines:
#         p_1 = line[0]
#         p_2 = line[1]
#         x_list.extend(
#             (p_1[0], p_2[0], None)
#         )
#         y_list.extend(
#             (p_1[1], p_2[1], None)
#         )
#         z_list.extend(
#             (lvl.elevation, lvl.elevation, None)
#         )
#     data_dict.append({
#         "type": "scatter3d",
#         "mode": "lines",
#         "x": x_list,
#         "y": y_list,
#         "z": z_list,
#         "hoverinfo": "skip",
#         "line": {
#             "width": 4,
#             "color": graphics_common.BISECTOR_COLOR
#         }
#     })


def add_data__frames(data_dict, mdl, load_case):
    """
    Adds a trace containing frame element centroidal axis lines
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
      load_case (LoadCase): the load_case to be visualized
    """
    beamcolumn_elems = mdl.list_of_beamcolumn_elements()
    if not beamcolumn_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    customdata_list = []
    section_names = []
    for elm in beamcolumn_elems:
        if elm.visibility.hidden_at_line_plots:
            continue
        p_i = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        p_j = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        section_names.extend([elm.section.name]*3)
        x_list.extend(
            (p_i[0], p_j[0], None)
        )
        y_list.extend(
            (p_i[1], p_j[1], None)
        )
        z_list.extend(
            (p_i[2], p_j[2], None)
        )
        if load_case:
            customdata_list.append(
                (elm.uid,
                 *load_case.line_element_udl[elm.uid].val,
                 elm.nodes[0].uid,
                 elm.parent_component.uid)
            )
            customdata_list.append(
                (elm.uid,
                 *load_case.line_element_udl[elm.uid].val,
                 elm.nodes[1].uid,
                 elm.parent_component.uid)
            )
            customdata_list.append(
                (None,)*6
            )
        else:
            customdata_list.append(
                (elm.uid,
                 elm.nodes[0].uid,
                 elm.parent_component.uid)
            )
            customdata_list.append(
                (elm.uid,
                 elm.nodes[1].uid,
                 elm.parent_component.uid)
            )
            customdata_list.append(
                (None,)*3
            )

    if load_case:
        customdata: nparr = np.array(customdata_list, dtype='object')
        data_dict.append({
            "name": "Frame elements",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "text": section_names,
            "customdata": customdata,
            "hovertemplate": 'Section: %{text}<br>' +
            'UDL (local): (%{customdata[1]:.3g}, ' +
            '%{customdata[2]:.3g}, %{customdata[3]:.3g})' +
            '<extra>Element: %{customdata[0]:d}<br>' +
            'Node @ this end: %{customdata[4]:d}<br>'
            'Parent: %{customdata[5]}</extra>',
            "line": {
                "width": 5,
                "color": graphics_common.FRAME_COLOR
            }
        })
    else:
        customdata = np.array(customdata_list, dtype='object')
        data_dict.append({
            "name": "Frame elements",
            "type": "scatter3d",
            "mode": "lines",
            "x": x_list,
            "y": y_list,
            "z": z_list,
            "text": section_names,
            "customdata": customdata,
            "hovertemplate": 'Section: %{text}<br>' +
            '<extra>Element: %{customdata[0]:d}<br>' +
            'Node @ this end: %{customdata[1]:d}<br>'
            'Parent: %{customdata[2]}</extra>',
            "line": {
                "width": 5,
                "color": graphics_common.FRAME_COLOR
            }
        })


def add_data__frame_offsets(data_dict, mdl):
    """
    Adds a trace containing frame element rigid offset lines
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
    """
    beamcolumn_elems = mdl.list_of_beamcolumn_elements()
    if not beamcolumn_elems:
        return

    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []

    for elm in beamcolumn_elems:
        p_i: nparr = np.array(elm.nodes[0].coords)
        p_io: nparr = (np.array(elm.nodes[0].coords)
                       + elm.geomtransf.offset_i)
        p_j: nparr = np.array(elm.nodes[1].coords)
        p_jo: nparr = (np.array(elm.nodes[1].coords)
                       + elm.geomtransf.offset_j)

        x_list.extend((p_i[0], p_io[0], None))
        y_list.extend((p_i[1], p_io[1], None))
        z_list.extend((p_i[2], p_io[2], None))
        x_list.extend((p_j[0], p_jo[0], None))
        y_list.extend((p_j[1], p_jo[1], None))
        z_list.extend((p_j[2], p_jo[2], None))

    data_dict.append({
        "name": "Rigid offsets",
        "type": "scatter3d",
        "mode": "lines",
        "x": x_list,
        "y": y_list,
        "z": z_list,
        "hoverinfo": "skip",
        "line": {
            "width": 8,
            "color": graphics_common.OFFSET_COLOR
        }
    })


def add_data__frame_axes(data_dict, mdl, ref_len):
    """
    Adds a trace containing frame element local axis lines
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
      ref_len (float): model reference length to scale the axes
    """
    beamcolumn_elems = mdl.list_of_beamcolumn_elements()
    if not beamcolumn_elems:
        return
    scaling = ref_len * 0.025
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    colors: list[Optional[str]] = []
    for elm in beamcolumn_elems:
        if elm.visibility.hidden_at_line_plots:
            continue
        x_vec = elm.geomtransf.x_axis
        y_vec = elm.geomtransf.y_axis
        z_vec = elm.geomtransf.z_axis
        l_clear = elm.clear_length()
        i_pos = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        mid_pos = i_pos + x_vec * l_clear/2.00
        x_list.extend((mid_pos[0], mid_pos[0]+x_vec[0]*scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1]+x_vec[1]*scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2]+x_vec[2]*scaling, None))
        colors.extend(["red"]*3)
        x_list.extend((mid_pos[0], mid_pos[0]+y_vec[0]*scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1]+y_vec[1]*scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2]+y_vec[2]*scaling, None))
        colors.extend(["green"]*3)
        x_list.extend((mid_pos[0], mid_pos[0]+z_vec[0]*scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1]+z_vec[1]*scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2]+z_vec[2]*scaling, None))
        colors.extend(["blue"]*3)
    data_dict.append({
        "name": "Frame axes",
        "type": "scatter3d",
        "mode": "lines",
        "x": x_list,
        "y": y_list,
        "z": z_list,
        "hoverinfo": "skip",
        "line": {
            "width": 8,
            "color": colors
        }
    })


def add_data__zerolength_axes(data_dict, mdl, ref_len):
    """
    Adds a trace containing zerolength element local axis lines
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
      ref_len (float): model reference length to scale the axes
    """
    zerolength_elements = mdl.list_of_zerolength_elements()
    if not zerolength_elements:
        return
    scaling = ref_len * 0.025
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    colors: list[Optional[str]] = []
    for elm in zerolength_elements:
        x_vec: nparr = elm.vecx
        y_vec: nparr = elm.vecyp
        z_vec: nparr = np.cross(x_vec, y_vec)
        mid_pos = np.array(elm.nodes[0].coords)
        x_list.extend((mid_pos[0], mid_pos[0]+x_vec[0]*scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1]+x_vec[1]*scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2]+x_vec[2]*scaling, None))
        colors.extend(["red"]*3)
        x_list.extend((mid_pos[0], mid_pos[0]+y_vec[0]*scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1]+y_vec[1]*scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2]+y_vec[2]*scaling, None))
        colors.extend(["green"]*3)
        x_list.extend((mid_pos[0], mid_pos[0]+z_vec[0]*scaling, None))
        y_list.extend((mid_pos[1], mid_pos[1]+z_vec[1]*scaling, None))
        z_list.extend((mid_pos[2], mid_pos[2]+z_vec[2]*scaling, None))
        colors.extend(["blue"]*3)
    data_dict.append({
        "name": "Zerolength axes",
        "type": "scatter3d",
        "mode": "lines",
        "x": x_list,
        "y": y_list,
        "z": z_list,
        "hoverinfo": "skip",
        "line": {
            "width": 8,
            "color": colors
        }
    })


def add_data__global_axes(data_dict, ref_len):
    """
    Adds a trace containing global axes
    Arguments:
      data_dict (dict): dictionary containing figure data
      ref_len (float): model reference length to scale the axes
    """
    scaling = ref_len
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    colors: list[Optional[str]] = []
    x_vec: nparr = np.array([1.00, 0.00, 0.00])
    y_vec: nparr = np.array([0.00, 1.00, 0.00])
    z_vec: nparr = np.array([0.00, 0.00, 1.00])

    x_list.extend((0.00, x_vec[0]*scaling, None))
    y_list.extend((0.00, x_vec[1]*scaling, None))
    z_list.extend((0.00, x_vec[2]*scaling, None))
    x_list.extend((0.00, y_vec[0]*scaling, None))
    y_list.extend((0.00, y_vec[1]*scaling, None))
    z_list.extend((0.00, y_vec[2]*scaling, None))
    x_list.extend((0.00, z_vec[0]*scaling, None))
    y_list.extend((0.00, z_vec[1]*scaling, None))
    z_list.extend((0.00, z_vec[2]*scaling, None))
    colors.extend(["red"]*3)
    colors.extend(["green"]*3)
    colors.extend(["blue"]*3)
    data_dict.append({
        "type": "scatter3d",
        "mode": "lines",
        "x": x_list,
        "y": y_list,
        "z": z_list,
        "hoverinfo": "skip",
        "line": {
            "width": 3,
            "color": colors
        }
    })


def add_data__extruded_frames_mesh(data_dict, mdl):
    """
    Adds a trace containing frame element extrusion mesh
    Arguments:
      data_dict (dict): dictionary containing figure data
      mdl (Model): the model to be visualized
    """
    beamcolumn_elems = mdl.list_of_beamcolumn_elements()
    if not beamcolumn_elems:
        return
    x_list: list[Optional[float]] = []
    y_list: list[Optional[float]] = []
    z_list: list[Optional[float]] = []
    i_list: list[Optional[int]] = []
    j_list: list[Optional[int]] = []
    k_list: list[Optional[int]] = []
    index = 0
    for elm in beamcolumn_elems:
        if elm.visibility.hidden_when_extruded:
            continue
        side_a = np.array(elm.nodes[0].coords) + elm.geomtransf.offset_i
        side_b = np.array(elm.nodes[1].coords) + elm.geomtransf.offset_j
        y_vec = elm.geomtransf.y_axis
        z_vec = elm.geomtransf.z_axis
        if not elm.section.outside_shape:
            continue
        loop = elm.section.outside_shape.halfedges
        for halfedge in loop:
            loc0 = halfedge.vertex.coords[0]*z_vec +\
                halfedge.vertex.coords[1]*y_vec + side_a
            loc1 = halfedge.vertex.coords[0]*z_vec +\
                halfedge.vertex.coords[1]*y_vec + side_b
            loc2 = halfedge.nxt.vertex.coords[0]*z_vec +\
                halfedge.nxt.vertex.coords[1]*y_vec + side_b
            loc3 = halfedge.nxt.vertex.coords[0]*z_vec +\
                halfedge.nxt.vertex.coords[1]*y_vec + side_a
            x_list.append(loc0[0])
            y_list.append(loc0[1])
            z_list.append(loc0[2])
            x_list.append(loc1[0])
            y_list.append(loc1[1])
            z_list.append(loc1[2])
            x_list.append(loc2[0])
            y_list.append(loc2[1])
            z_list.append(loc2[2])
            x_list.append(loc3[0])
            y_list.append(loc3[1])
            z_list.append(loc3[2])
            i_list.append(index + 0)
            j_list.append(index + 1)
            k_list.append(index + 2)
            i_list.append(index + 0)
            j_list.append(index + 2)
            k_list.append(index + 3)
            index += 4
    data_dict.append({
        "type": "mesh3d",
        "x": x_list,
        "y": y_list,
        "z": z_list,
        "i": i_list,
        "j": j_list,
        "k": k_list,
        "hoverinfo": "skip",
        "color": graphics_common.BEAM_MESH_COLOR,
        "opacity": 0.30
    })


def show(mdl: Model,
         load_case: Optional[LoadCase] = None,
         extrude=False,
         offsets=True,
         global_axes=True,
         diaphragm_lines=True,
         tributary_areas=True,
         parent_nodes=True,
         frame_axes=False,
         zerolength_axes=False,
         camera=None
         ):
    """
    Visualize the model
    Arguments:
      mdl (Model): the model to be visualized
      load_case (LoadCase): the load_case to be visualized
      extrude (bool): wether to extrude frame elements
      offsets (bool): whether to show frame element rigid offsets
      global_axes (bool): whether to show global axes
      diaphragm_lines (bool): whether to show lines indicating rigid
        diaphragm extent
      tributary_areas (bool): whether to show tributary area
        boundaries
      parent_nodes (bool): whether to plot parent nodes
      frame_axes (bool): whether to show the local axes of frame
        elements
      zerolength_axes (bool): whether to show the local axes of
        zerolength elements
      camera (dict): custom positioning of the camera
    """

    layout = graphics_common_3d.global_layout(camera)
    data_dict: list[dict[str, object]] = []

    ref_len = mdl.reference_length()

    # plot the nodes

    add_data__nodes(data_dict, mdl, load_case)
    if not extrude:
        add_data__internal_nodes(data_dict, mdl, load_case)

    # global axes
    if global_axes:
        add_data__global_axes(data_dict, ref_len)

    # # diaphgragm lines
    # if diaphragm_lines:
    #     for lvl in mdl.levels.values():
    #         add_data__diaphragm_lines(data_dict, lvl)

    # # bisector lines
    # if tributary_areas:
    #     for lvl in mdl.levels.values():
    #         add_data__bisector_lines(data_dict, lvl)

    # plot beamcolumn elements
    add_data__frames(data_dict, mdl, load_case)
    if frame_axes:
        add_data__frame_axes(data_dict, mdl, ref_len)
    if zerolength_axes:
        add_data__zerolength_axes(data_dict, mdl, ref_len)
    if extrude:
        add_data__extruded_frames_mesh(
            data_dict, mdl)
    # plot the rigid offsets
    if offsets:
        add_data__frame_offsets(data_dict, mdl)

    # plot the parent nodes
    if parent_nodes:
        if load_case:
            add_data__parent_nodes(data_dict, load_case)

    fig_datastructure = dict(data=data_dict, layout=layout)
    fig = go.Figure(fig_datastructure)

    if "pytest" not in sys.modules:
        fig.show()



"""
Model Generator for OpenSees ~ rigid diaphragm
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import TYPE_CHECKING
from dataclasses import dataclass
import numpy as np
from ..ops.node import Node
from .. import common
from .. import load_case
if TYPE_CHECKING:
    from ..load_case import LoadCase
    from ..level import Level


@dataclass(repr=False)
class RDAnalyzer:
    """
    Rigid Diaphragm Analyzer object.
    """
    parent_loadcase: LoadCase
    parent_level: Level

    def run(self, gather_mass):
        """
        Applies rigid diaphragm constraints and optionally
        gathers mass from the nodes of the level to the
        primary nodes
        """
        lvl = self.parent_level
        loadcase = self.parent_loadcase
        # gather all nodes of the level
        nodes: list[Node] = []
        nodes.extend(lvl.nodes.values())
        for component in lvl.components.values():
            nodes.extend(component.internal_nodes.values())

        # determine center of mass
        masses = []
        coords = []
        for node in nodes:
            mass = loadcase.node_mass[node.uid].val[0]
            masses.append(mass)
            coords.append(node.coords[0:2])
        total_mass = sum(masses)
        if np.abs(total_mass) <= common.EPSILON:
            raise ValueError(
                "Can't generate parent node without defined mass.")
        coords_np = np.array(coords) * np.column_stack((masses, masses))
        center = np.sum(coords_np/total_mass, axis=0)
        parent_node = Node(
            lvl.parent_model.uid_generator.new('node'),
            [*center, lvl.elevation])
        loadcase.parent_nodes[lvl.uid] = parent_node
        loadcase.node_loads[parent_node.uid] = \
            load_case.PointLoadMass()
        loadcase.node_mass[parent_node.uid] = \
            load_case.PointLoadMass()
        if gather_mass:
            # gather all mass from the nodes of the level to the
            # parent node
            level_nodes: list[Node] = []
            level_nodes.extend(lvl.nodes.values())
            for comp in lvl.components.values():
                level_nodes.extend(comp.internal_nodes.values())
            for node in level_nodes:
                mass = loadcase.node_mass[node.uid].val[0]
                loadcase.node_mass[node.uid].val = np.zeros(6)
                dist2 = np.linalg.norm(
                    np.array(node.coords[0:2])
                    - np.array(parent_node.coords[0:2]))**2
                loadcase.node_mass[parent_node.uid].val += \
                    np.array((mass, mass, 0.00, 0.00, 0.00, mass*dist2))



"""
Enables functionality by utilizing the `halfedge`
data structure.
"""

#                          __
#   ____  ____ ___  ____ _/ /
#  / __ \/ __ `__ \/ __ `/ /
# / /_/ / / / / / / /_/ /_/
# \____/_/ /_/ /_/\__, (_)
#                /____/
#
# https://github.com/ioannis-vm/OpenSees_Model_Generator

from __future__ import annotations
from typing import Optional
from typing import Any
from itertools import count
from descartes.patch import PolygonPatch  # type: ignore
import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt   # type: ignore
from shapely.geometry import Polygon as shapely_Polygon   # type: ignore
from .import common

nparr = npt.NDArray[np.float64]


class Vertex:
    """
    2D Vertex.
    It knows all the edges connected to it.
    It knows all the halfedges leaving from it.
    Each instance has an automatically generated unique id.
    """
    _ids = count(0)

    def __init__(self, coords: tuple[float, float]):
        self.coords = coords
        self.edges: list[Edge] = []
        self.halfedges: list[Halfedge] = []
        self.uid: int = next(self._ids)

    def __eq__(self, other):
        return self.uid == other.uid

    def __repr__(self):
        return f'(V{self.uid} @ {self.coords}) '


class Edge:
    """
    2D oriented Edge.
    Connected to two vertices v_i and v_j.
    Has two halfedges, h_i and h_j.
    Each instance has an automatically generated unique id.
    """

    _ids = count(0)

    def __init__(self, v_i: Vertex, v_j: Vertex):
        self.v_i = v_i
        self.v_j = v_j
        self.uid = next(self._ids)
        self.h_i: Optional[Halfedge] = None
        self.h_j: Optional[Halfedge] = None
        if self not in self.v_i.edges:
            self.v_i.edges.append(self)
        if self not in self.v_j.edges:
            self.v_j.edges.append(self)

    def __repr__(self):
        return f'(E{self.uid} @ V{self.v_i.uid}, V{self.v_j.uid}) '

    def define_halfedge(self, vertex: Vertex):
        """
        For the current edge instance and given one of its vertices,
        we want the halfedge that points to the direction
        away from the given vertex.
        We create it if it does not exist.
        """
        if vertex == self.v_i:
            if not self.h_i:
                halfedge = Halfedge(self.v_i, self)
                self.h_i = halfedge
            else:
                raise ValueError('Halfedge h_i already defined')
        elif vertex == self.v_j:
            if not self.h_j:
                halfedge = Halfedge(self.v_j, self)
                self.h_j = halfedge
            else:
                raise ValueError('Halfedge h_j already defined')
        else:
            raise ValueError(
                "The edge is not connected to the given vertex.")
        return halfedge

    def other_vertex(self, vertex):
        """
        We have an edge instance.
        It has two vertices. This method returns
        the other vertex provided one of the vertices.
        """
        if self.v_i == vertex:
            v_other = self.v_j
        elif self.v_j == vertex:
            v_other = self.v_i
        else:
            raise ValueError("The edge is not connected to the given vertex")
        return v_other


class Halfedge:
    """
    Halfedge object.
    Every edge has two halfedges.
    A halfedge has a direction, pointing from one
    of the corresponding edge's vertices to the other.
    The `vertex` attribute corresponds to the
    edge's vertex that the halfedge originates from.
    Halfedges have a `next` attribute that
    points to the next halfedge, forming closed
    loops, or sequences, which is the purpose of this module.
    """

    _ids = count(0)

    def __init__(self, vertex: Vertex, edge: Edge):
        self.vertex = vertex
        self.edge = edge
        self.uid: int = next(self._ids)
        self.nxt = None

    def __repr__(self):
        if self.nxt:
            out = f'(H{self.uid} from E{self.edge.uid}' \
                f' to E{self.nxt.edge.uid} next H{self.nxt.uid})'
        else:
            out = f'(H{self.uid}'
        return out

    def __lt__(self, other):
        return self.uid < other.uid

    def direction(self):
        """
        Calculates the angular direction of the halfedge
        using the arctan2 function
        """
        drct: nparr = (np.array(self.edge.other_vertex(self.vertex).coords) -
                       np.array(self.vertex.coords))
        norm = np.linalg.norm(drct)
        drct /= norm
        return np.arctan2(drct[1], drct[0])


class Mesh:
    """
    A container that holds a list of unique halfedges.
    Vertices and edges can be retrieved from those.
    The mesh is assumed to be flat (2D).
    """

    def __init__(self, halfedges: list[Halfedge]):
        self.halfedges = halfedges

    def __repr__(self):
        num = len(self.halfedges)
        return f'Mesh object containing {num} halfedges.'

    def geometric_properties(self):
        """
        Calculates the geometric properties of the shape defined by
        the mesh
        """
        coords: nparr = np.array([h.vertex.coords for h in self.halfedges])
        return geometric_properties(coords)

    def bounding_box(self):
        """
        Returns a bounding box of the mesh
        """
        coords: nparr = np.array([h.vertex.coords for h in self.halfedges])
        xmin = min(coords[:, 0])
        xmax = max(coords[:, 0])
        ymin = min(coords[:, 1])
        ymax = max(coords[:, 1])
        return np.array([[xmin, ymin], [xmax, ymax]])


############################################
# Geometric Properties of Polygonal Shapes #
############################################


def polygon_area(coords: nparr) -> float:
    """
    Calculates the area of a polygon.
    Args:
        coords: A matrix whose columns represent
                the coordinates and the rows
                represent the points of the polygon.
                The first point should not be repeated
                at the end, as this is done
                automatically.
    Returns:
        area (float): The area of the polygon.
    """
    x_coords = coords[:, 0]
    y_coords = coords[:, 1]
    return float(np.sum(x_coords * np.roll(y_coords, -1) -
                        np.roll(x_coords, -1) * y_coords) / 2.00)


def polygon_centroid(coords: nparr) -> nparr:
    """
    Calculates the centroid of a polygon.
    Args:
        coords: A matrix whose columns represent
                the coordinates and the rows
                represent the points of the polygon.
                The first point should not be repeated
                at the end, as this is done
                automatically.
    Returns:
        centroid (nparr): The centroid of
                 the polygon.
    """
    x_coords = coords[:, 0]
    y_coords = coords[:, 1]
    area = polygon_area(coords)
    x_cent = (np.sum((x_coords + np.roll(x_coords, -1)) *
                     (x_coords*np.roll(y_coords, -1) -
                      np.roll(x_coords, -1)*y_coords)))/(6.0*area)
    y_cent = (np.sum((y_coords + np.roll(y_coords, -1)) *
                     (x_coords*np.roll(y_coords, -1) -
                      np.roll(x_coords, -1)*y_coords)))/(6.0*area)
    return np.array((x_cent, y_cent))


def polygon_inertia(coords):
    """
    Calculates the moments of inertia of a polygon.
    Args:
        coords: A matrix whose columns represent
                the coordinates and the rows
                represent the points of the polygon.
                The first point should not be repeated
                at the end, as this is done
                automatically.
    Returns:
        dictionary, containing:
        'ixx': (float) - Moment of inertia around
                         the x axis
        'iyy': (float) - Moment of inertia around
                         the y axis
        'ixy': (float) - Product of inertia
        'ir': (float)  - Polar moment of inertia
        'ir_mass': (float) - Mass moment of inertia
        # TODO
        # The terms might not be pedantically accurate
    """
    x_coords = coords[:, 0]
    y_coords = coords[:, 1]
    area = polygon_area(coords)
    alpha = x_coords * np.roll(y_coords, -1) - \
        np.roll(x_coords, -1) * y_coords
    # planar moment of inertia wrt horizontal axis
    ixx = np.sum((y_coords**2 + y_coords * np.roll(y_coords, -1) +
                  np.roll(y_coords, -1)**2)*alpha)/12.00
    # planar moment of inertia wrt vertical axis
    iyy = np.sum((x_coords**2 + x_coords * np.roll(x_coords, -1) +
                  np.roll(x_coords, -1)**2)*alpha)/12.00

    ixy = np.sum((x_coords*np.roll(y_coords, -1)
                  + 2.0*x_coords*y_coords
                  + 2.0*np.roll(x_coords, -1) * np.roll(y_coords, -1)
                  + np.roll(x_coords, -1) * y_coords)*alpha)/24.
    # polar (torsional) moment of inertia
    i_r = ixx + iyy
    # mass moment of inertia wrt in-plane rotation
    ir_mass = (ixx + iyy) / area

    return {'ixx': ixx, 'iyy': iyy,
            'ixy': ixy, 'ir': i_r, 'ir_mass': ir_mass}


def geometric_properties(coords):
    """
    Aggregates the results of the previous functions.
    """

    # repeat the first row at the end to close the shape
    coords = np.vstack((coords, coords[0, :]))
    area = polygon_area(coords)
    centroid = polygon_centroid(coords)
    coords_centered = coords - centroid
    inertia = polygon_inertia(coords_centered)

    return {'area': area, 'centroid': centroid, 'inertia': inertia}


##################################
# Defining halfedges given edges #
##################################

# auxiliary functions

def ang_reduce(ang):
    """
    Brings and angle expressed in radians in the interval [0, 2pi)
    """
    while ang < 0:
        ang += 2.*np.pi
    while ang >= 2.*np.pi:
        ang -= 2.*np.pi
    return ang


def define_halfedges(edges: list[Edge]) -> list[Halfedge]:
    """
    Given a list of edges, defines all the halfedges and
    associates them with their `next`.
    See note:
        https://notability.com/n/0wlJ17mt81uuVWAYVoFfV3
    To understand how it works.
    Description:
        Each halfedge stores information about its edge, vertex
        and and next halfedge. Contrary to convention, we don't
        store the twin (opposite) halfedge here, seince we don't
        need it anywhere.
    Args:
        edges (list[Edge]): List of Edge objects
    Returns:
        halfedges (list[Halfedge]): List of Halfedge objects
    """

    all_halfedges = []
    for edge in edges:
        v_i = edge.v_i
        v_j = edge.v_j
        h_i = edge.define_halfedge(v_i)
        h_j = edge.define_halfedge(v_j)
        all_halfedges.append(h_i)
        all_halfedges.append(h_j)
        v_i.halfedges.append(h_i)
        v_j.halfedges.append(h_j)

    # at this point we have defined all halfedges, but
    # none of them knows its `next`.
    # We now assign that attribute to all halfedges

    for halfedge in all_halfedges:
        # We are looking for `h`'s `next`
        # determine the vertex that it starts from
        v_from = halfedge.vertex
        # determine the vertex that it points to
        v_to = halfedge.edge.other_vertex(v_from)
        # get a list of all halfedges leaving that vertex
        candidates_for_next = v_to.halfedges
        # determine which of all these halfedges will be the next
        angles = np.full(len(candidates_for_next), 0.00)
        for i, h_other in enumerate(candidates_for_next):
            if h_other.edge == halfedge.edge:
                angles[i] = 1000.
                # otherwise we would assign its conjugate as next
            else:
                angles[i] = ang_reduce(
                    (halfedge.direction() - np.pi) - h_other.direction())
        halfedge.nxt = candidates_for_next[np.argmin(angles)]

    return all_halfedges

    # # debug
    # import matplotlib.pyplot as plt
    # fig = plt.figure()
    # ax = fig.add_subplot(111)
    # ax.set_aspect('equal')
    # for edge in edges:
    #     p1 = edge.v_i.coords
    #     p2 = edge.v_j.coords
    #     coords = np.row_stack((p1, p2))
    #     ax.plot(coords[:, 0], coords[:, 1])
    # for h in halfedges:
    #     if h.nxt:
    #         h_nxt = h.nxt
    #         e = h.edge
    #         if h_nxt.edge:
    #             e_nxt = h_nxt.edge
    #             p1 = (np.array(e.v_i.coords)
    #                   + np.array(e.v_j.coords))/2.
    #             p2 = (np.array(e_nxt.v_i.coords)
    #                   + np.array(e_nxt.v_j.coords))/2.
    #             dx = p2 - p1
    #             ax.arrow(*p1, *dx)
    # plt.show()


def obtain_closed_loops(halfedges):
    """
    Given a list of halfedges,
    this function uses their `next` attribute to
    group them into sequences of closed loops
    (ordered lists of halfedges of which the
    `next` halfedge of the last list element
    points to the first halfedge in the list, and
    the `next` halfedge of any list element
    points to the next halfedge in the list.
    Args:
        halfedges (list[Halfedge]):
                  list of halfedges
    Returns:
        loops (list[list[Halfedge]]) with the
              aforementioned property.
    """
    def is_in_some_loop(halfedge, loops):
        for loop in loops:
            if halfedge in loop:
                return True
        return False
    loops: list[list[Halfedge]] = []
    for halfedge in halfedges:
        if loops:
            if is_in_some_loop(halfedge, loops):
                continue
        loop = [halfedge]
        nxt = halfedge.nxt
        while nxt != halfedge:
            loop.append(nxt)
            nxt = nxt.nxt
        loops.append(loop)
    return loops


def orient_loops(loops):
    """
    Separates loops to internal (counterclockwise)
    and external (clockwise). Also gathers trivial
    loops, i.e. halfedge sequences that define polygons
    that have no area (e.g. h1 -> h2 -> h1).
    Args:
        loops (list[list[Halfedge]]) (see `obtain_closed_loops`)
    Returns:
        external_loops (list[list[Halfedge]])
        internal_loops (list[list[Halfedge]])
        trivial_loops (list[list[Halfedge]])
    """
    internal_loops = []
    external_loops = []
    trivial_loops = []
    loop_areas = [polygon_area(
        np.array([h.vertex.coords for h in loop]))
        for loop in loops]
    for i, area in enumerate(loop_areas):
        if area > common.EPSILON:
            internal_loops.append(loops[i])
        elif area < -common.EPSILON:
            external_loops.append(loops[i])
        else:
            trivial_loops.append(loops[i])
    return external_loops, internal_loops, trivial_loops


#######################################
# Breaking a shape into little pieces #
#######################################


def subdivide_polygon(outside, holes, n_x, n_y, plot=False):
    """
    Used to define the fibers of fiber sections.
    Args:
        halfedges (list[Halfedge]): Sequence of halfedges
                  that defines the shape of a section.
        n_x (int): Number of spatial partitions in the x direction
        n_y (int): Number of spatial partitions in the y direction
        plot (bool): Plots the resulting polygons for debugging
    Returns:
        pieces (list[shapely_Polygon]): shapely_Polygon
               objects that represent single fibers.
    """
    outside_polygon = shapely_Polygon(
        [h.vertex.coords for h in outside.halfedges])
    hole_polygons = []
    for hole in holes.values():
        hole_polygons.append(shapely_Polygon(
            [h.vertex.coords for h in hole.halfedges]))
    remaining_polygon = outside_polygon
    for hole_polygon in hole_polygons:
        remaining_polygon = remaining_polygon.difference(hole_polygon)
    x_min, y_min, x_max, y_max = outside_polygon.bounds
    x_array = np.linspace(x_min, x_max, num=n_x, endpoint=True)
    y_array = np.linspace(y_min, y_max, num=n_y, endpoint=True)
    pieces = []
    for i in range(len(x_array)-1):
        for j in range(len(y_array)-1):
            tile = shapely_Polygon([(x_array[i], y_array[j]),
                                    (x_array[i+1], y_array[j]),
                                    (x_array[i+1], y_array[j+1]),
                                    (x_array[i], y_array[j+1])])
            subregion = remaining_polygon.intersection(tile)
            if subregion.area != 0.0:
                pieces.append(subregion)
    if plot:
        fig = plt.figure()
        ax_1 = fig.add_subplot(111)
        ax_1.set_aspect('equal')
        patch = PolygonPatch(remaining_polygon, alpha=0.5, zorder=2)
        ax_1.add_patch(patch)
        for subregion in pieces:
            patch = PolygonPatch(subregion, alpha=0.5, zorder=2)
            ax_1.add_patch(patch)
        for subregion in pieces:
            ax_1.scatter(subregion.centroid.x, subregion.centroid.y)
        ax_1.margins(0.10)
        plt.show()
    return pieces


#############
# Debugging #
#############


def print_halfedge_results(halfedges):
    """
    Prints the ids of the defined halfedges
    and their vertex, edge and next, for
    debugging.
    """
    results: dict[str, list[Any]] = {
        'halfedge': [],
        'vertex': [],
        'edge': [],
        'next': [],
    }

    for halfedge in halfedges:
        results['halfedge'].append(halfedge)
        results['vertex'].append(halfedge.vertex)
        results['edge'].append(halfedge.edge)
        results['next'].append(halfedge.nxt)

    print(results)


def plot_loop(halfedge_loop):
    """
    Plots the vertices/edges of a list of halfedges.
    """
    num = len(halfedge_loop)
    coords = np.full((num+1, 2), 0.00)
    for i, halfedge in enumerate(halfedge_loop):
        coords[i, :] = halfedge.vertex.coords
    coords[-1, :] = coords[0, :]
    fig = plt.figure()
    plt.plot(coords[:, 0], coords[:, 1])
    plt.scatter(coords[:, 0], coords[:, 1])
    fig.show()


def plot_edges(edges):
    """
    Plots the given edges.
    """
    fig = plt.figure()
    for edge in edges:
        coords = np.full((2, 2), 0.00)
        coords[0, :] = edge.v_i.coords
        coords[1, :] = edge.v_j.coords
        plt.plot(coords[:, 0], coords[:, 1])
    fig.show()


def sanity_checks(external, trivial):
    """
    Perform some checks to make sure
    assumptions are not violated.
    """
    #   We expect no trivial loops
    if trivial:
        print("Warning: Found trivial loop")
        for i, trv in enumerate(trivial):
            for halfedge in trv:
                print(halfedge.vertex.coords)
            plot_loop(trv)
    #   We expect a single external loop
    if len(external) > 1:
        print("Warning: Found multiple external loops")
        for i, ext in enumerate(external):
            print(i+1)
            for halfedge in ext:
                print(halfedge.vertex.coords)
            plot_loop(ext)


if __name__ == "__main()__":
    pass
